import numpy as np
import readsnap
import readsubf
import sys
import time
import random

###############################################################################
#this function returns an array containing the positions of the galaxies (kpc/h)
#in the catalogue according to the fiducial density, M1 and alpha
#CDM halos with masses within [min_mass,max_mass], are populated
#with galaxies. The IDs and positions of the CDM particles belonging to the
#different groups are read from the snapshots
#If one needs to creates many catalogues, this function is not appropiate,
#since it wastes a lot of time reading the snapshots and sorting the IDs
#min_mass and max_mass are in units of Msun/h, not 1e10 Msun/h
#mass_criteria: definition of the halo virial radius -- 't200' 'm200' 'c200'
#fiducial_density: galaxy number density to be reproduced, in (h/Mpc)^3
def hod(snapshot_fname,groups_fname,groups_number,min_mass,max_mass,
        fiducial_density,M1,alpha,mass_criteria,verbose=False):

    thres=1e-3 #controls the max relative error to accept a galaxy density
    
    #read the header and obtain the boxsize
    head=readsnap.snapshot_header(snapshot_fname)
    BoxSize=head.boxsize    #BoxSize in kpc/h

    #read positions and IDs of DM particles: sort the IDs array
    DM_pos=readsnap.read_block(snapshot_fname,"POS ",parttype=-1) #kpc/h
    DM_ids=readsnap.read_block(snapshot_fname,"ID  ",parttype=-1)-1 
    sorted_ids=DM_ids.argsort(axis=0)
    #the particle whose ID is N is located in the position sorted_ids[N]
    #i.e. DM_ids[sorted_ids[N]]=N
    #the position of the particle whose ID is N would be:
    #DM_pos[sorted_ids[N]]

    #read the IDs of the particles belonging to the CDM halos
    halos_ID=readsubf.subf_ids(groups_fname,groups_number,0,0,
                               long_ids=True,read_all=True)
    IDs=halos_ID.SubIDs-1
    del halos_ID

    #read CDM halos information
    halos=readsubf.subfind_catalog(groups_fname,groups_number,
                                   group_veldisp=True,masstab=True,
                                   long_ids=True,swap=False)
    if mass_criteria=='t200':
        halos_mass=halos.group_m_tophat200*1e10   #masses in Msun/h
        halos_radius=halos.group_r_tophat200      #radius in kpc/h
    elif mass_criteria=='m200':
        halos_mass=halos.group_m_mean200*1e10     #masses in Msun/h
        halos_radius=halos.group_r_mean200        #radius in kpc/h
    elif mass_criteria=='c200':    
        halos_mass=halos.group_m_crit200*1e10     #masses in Msun/h
        halos_radius=halos.group_r_crit200        #radius in kpc/h
    else:
        print('bad mass_criteria')
        sys.exit()
    halos_pos=halos.group_pos   #positions in kpc/h
    halos_len=halos.group_len
    halos_offset=halos.group_offset
    halos_indexes=np.where((halos_mass>min_mass) & (halos_mass<max_mass))[0]
    del halos
    
    if verbose:
        print(' ')
        print('total halos found=',halos_pos.shape[0])
        print('halos number density=',len(halos_pos)/(BoxSize*1e-3)**3)

    #keep only the halos in the given mass range 
    halo_mass=halos_mass[halos_indexes]
    halo_pos=halos_pos[halos_indexes]
    halo_radius=halos_radius[halos_indexes]
    halo_len=halos_len[halos_indexes]
    halo_offset=halos_offset[halos_indexes]
    del halos_indexes

    ##### COMPUTE Mmin GIVEN M1 & alpha #####
    i=0; max_iterations=20 #maximum number of iterations
    Mmin1=min_mass; Mmin2=max_mass
    while (i<max_iterations):
        Mmin=0.5*(Mmin1+Mmin2) #estimation of the HOD parameter Mmin

        total_galaxies=0
        inside=np.where(halo_mass>Mmin)[0] #take all galaxies with M>Mmin
        mass=halo_mass[inside] #only halos with M>Mmin have central/satellites

        total_galaxies=mass.shape[0]+np.sum((mass/M1)**alpha)
        mean_density=total_galaxies*1.0/(BoxSize*1e-3)**3 #galaxies/(Mpc/h)^3

        if (np.absolute((mean_density-fiducial_density)/fiducial_density)<thres):
            i=max_iterations
        elif (mean_density>fiducial_density):
            Mmin1=Mmin
        else:
            Mmin2=Mmin
        i+=1

    if verbose:
        print(' ')
        print('Mmin=',Mmin)
        print('average number of galaxies=',total_galaxies)
        print('average galaxy density=',mean_density)
    #########################################

    #just halos with M>Mmin; the rest do not host central/satellite galaxies
    inside=np.where(halo_mass>Mmin)[0]
    halo_mass=halo_mass[inside]
    halo_pos=halo_pos[inside]
    halo_radius=halo_radius[inside]
    halo_len=halo_len[inside]
    halo_offset=halo_offset[inside]
    del inside

    #compute number of satellites in each halo using the Poisson distribution 
    N_mean_sat=(halo_mass/M1)**alpha #mean number of satellites
    N_sat=np.empty(len(N_mean_sat),dtype=np.int32)
    for i in range(len(N_sat)):
        N_sat[i]=np.random.poisson(N_mean_sat[i])
    N_tot=np.sum(N_sat)+len(halo_mass) #total number of galaxies in the catalogue

    if verbose:
        print(' ')
        print(np.min(halo_mass),'< M_halo <',np.max(halo_mass))
        print('total number of galaxies=',N_tot)
        print('galaxy number density=',N_tot/(BoxSize*1e-3)**3)

    #put satellites following the distribution of dark matter in groups
    if verbose:
        print(' ')
        print('Creating mock catalogue ...',)

    pos_galaxies=np.empty((N_tot,3),dtype=np.float32)
    #index: variable that go through halos (may be several galaxies in a halo)
    #i: variable that go through all (central/satellites) galaxies
    #count: find number of galaxies that lie beyond its host halo virial radius
    index=0; count=0; i=0 
    while (index<halo_mass.shape[0]):

        position=halo_pos[index]  #position of the DM halo
        radius=halo_radius[index] #radius of the DM halo

        #save the position of the central galaxy
        pos_galaxies[i]=position; i+=1

        #if halo contains satellites, save their positions
        Nsat=N_sat[index]
        if Nsat>0:
            offset=halo_offset[index] 
            length=halo_len[index]
            idss=sorted_ids[IDs[offset:offset+length]]

            #compute the distances to the halo center keeping those with R<Rvir
            pos=DM_pos[idss] #positions of the particles belonging to the halo
            posc=pos-position

            #this is to populate correctly halos closer to box boundaries
            if np.any((position+radius>BoxSize) + (position-radius<0.0)):
                
                inside=np.where(posc[:,0]>BoxSize/2.0)[0]
                posc[inside,0]-=BoxSize
                inside=np.where(posc[:,0]<-BoxSize/2.0)[0]
                posc[inside,0]+=BoxSize

                inside=np.where(posc[:,1]>BoxSize/2.0)[0]
                posc[inside,1]-=BoxSize
                inside=np.where(posc[:,1]<-BoxSize/2.0)[0]
                posc[inside,1]+=BoxSize

                inside=np.where(posc[:,2]>BoxSize/2.0)[0]
                posc[inside,2]-=BoxSize
                inside=np.where(posc[:,2]<-BoxSize/2.0)[0]
                posc[inside,2]+=BoxSize
                
            radii=np.sqrt(posc[:,0]**2+posc[:,1]**2+posc[:,2]**2)
            inside=np.where(radii<radius)[0]
            selected=random.sample(inside,Nsat)
            pos=pos[selected]

            #aditional, not esential check. Can be comment out
            posc=pos-position
            if np.any((posc>BoxSize/2.0) + (posc<-BoxSize/2.0)):
                inside=np.where(posc[:,0]>BoxSize/2.0)[0]
                posc[inside,0]-=BoxSize
                inside=np.where(posc[:,0]<-BoxSize/2.0)[0]
                posc[inside,0]+=BoxSize

                inside=np.where(posc[:,1]>BoxSize/2.0)[0]
                posc[inside,1]-=BoxSize
                inside=np.where(posc[:,1]<-BoxSize/2.0)[0]
                posc[inside,1]+=BoxSize

                inside=np.where(posc[:,2]>BoxSize/2.0)[0]
                posc[inside,2]-=BoxSize
                inside=np.where(posc[:,2]<-BoxSize/2.0)[0]
                posc[inside,2]+=BoxSize
            r_max=np.max(np.sqrt(posc[:,0]**2+posc[:,1]**2+posc[:,2]**2))
            if r_max>radius: #check no particles beyond Rv selected 
                print(position)
                print(radius)
                print(pos)
                count+=1

            for j in range(Nsat):
                pos_galaxies[i]=pos[j]; i+=1
        index+=1

    if verbose:
        print('done')
    #some final checks
    if i!=N_tot:
        print('some galaxies missing:')
        print('register',i,'galaxies out of',N_tot)
    if count>0:
        print('error:',count,'particles beyond the virial radius selected')

    return pos_galaxies
###############################################################################





#This function is equal to the above one, except that the snapshot read, halos
#read and ID sorting it is not performing here. It is best suited when many 
#galaxy catalogues need to be created: for example, when iterating among M1 and 
#alpha trying to find the best combination that reproduces the measured wp(r)
#VARIABLES:
#DM_pos: array containing the positions of the CDM particles
#DM_vel: array containing the velocities of the CDM particles
#sorted_ids: array containing the positions of the IDs in the snapshots. 
#sorted_ids[N] gives the position where the particle whose ID is N is located
#IDs:IDs array as read from the subfind ID file
#halo_mass: array containing the masses of the CDM halos in the mass interval
#halo_pos:  array containing the positions of the CDM halos in the mass interval
#halo_vel:  array with the velocities of the CDM halos in the mass interval
#halo_radius: array containing the radii of the CDM halos in the mass interval
#halo_len: array containing the len of the CDM halos in the mass interval
#halo_offset: array containing the offset of the CDM halos in the mass interval
#BoxSize: Size of the simulation Box. In Mpc/h
#fiducial_density: galaxy number density to be reproduced, in (h/Mpc)^3
#model: 'standard' or 'LBG'. The former in the normal one whereas the latter is
#for the LBG HOD model which do not assign central galaxies to the halos
#NEW!!! In the latest version this routine requires both the positions and 
#velocities of the CDM particles. It will also return the galaxies velocities
class hod_fast:
    def __init__(self,DM_pos,DM_vel,sorted_ids,IDs,halo_mass,halo_pos,halo_vel,
                 halo_radius,halo_len,halo_offset,BoxSize,min_mass,max_mass,
                 fiducial_density,M1,alpha,seed,model='standard',
                 verbose=False):

        problematic_cases=0 #number of problematic cases (halos with Rvir=0 ...)
        thres=1e-3 #controls the max relative error to accept a galaxy density

        ##### COMPUTE Mmin GIVEN M1 & alpha #####
        i=0; max_iterations=20 #maximum number of iterations
        Mmin1=min_mass; Mmin2=max_mass
        while (i<max_iterations):
            #estimation of the HOD parameter Mmin
            Mmin=10**(0.5*(np.log10(Mmin1)+np.log10(Mmin2))) 

            total_galaxies=0

            #only halos with M>Mmin have central/satellites
            inside=np.where(halo_mass>Mmin)[0]; mass=halo_mass[inside] 
            
            if model=='standard': #central + satellites
                total_galaxies=mass.shape[0]+np.sum((mass/M1)**alpha)
            elif model=='LBG':    #only satellites
                total_galaxies=np.sum((mass/M1)**alpha)
            else:
                print('incorrect model'); sys.exit()
            mean_density=total_galaxies*1.0/BoxSize**3

            if (np.absolute((mean_density-fiducial_density)/fiducial_density)\
                    <thres):
                i=max_iterations
            elif (mean_density>fiducial_density):
                Mmin1=Mmin
            else:
                Mmin2=Mmin
            i+=1

        self.Mmin=Mmin
        if verbose:
            print('\nMmin = %.3e'%Mmin);
            print('average number of galaxies =',total_galaxies)
            print('average galaxy density = %.3e galaxies/(Mpc/h)^3'\
                %mean_density)
                
        #########################################

        #just halos with M>Mmin; the rest do not host central/satellite galaxies
        inside=np.where(halo_mass>Mmin)[0]
        halo_mass   = halo_mass[inside];   halo_pos = halo_pos[inside]
        halo_radius = halo_radius[inside]; halo_len = halo_len[inside]
        halo_offset = halo_offset[inside]; del inside

        #compute the # of satellites in each halo with the Poisson distribution 
        np.random.seed(seed) #this is just to check convergence on w_p(r_p)
        N_mean_sat=(halo_mass/M1)**alpha #mean number of satellites
        N_sat=np.empty(len(N_mean_sat),dtype=np.int32)
        for i in range(len(N_sat)):
            N_sat[i]=np.random.poisson(N_mean_sat[i])

        #total number of galaxies in the catalogue
        if model=='standard': #central + satellites
            N_tot=np.sum(N_sat)+len(halo_mass) 
        elif model=='LBG':    #only satellites
            N_tot=np.sum(N_sat) 

        self.galaxy_density=N_tot*1.0/BoxSize**3
        if verbose:
            print('\n%.3e < M_halo < %.3e'%(np.min(halo_mass),np.max(halo_mass)))
            print('total number of galaxies =',N_tot)
            print('galaxy number density = %.3e'%(N_tot/BoxSize**3))

        #put satellites following the distribution of dark matter in groups
        if verbose:
            print('\nCreating mock catalogue ...',)

        pos_galaxies=np.empty((N_tot,3),dtype=np.float32)
        vel_galaxies=np.empty((N_tot,3),dtype=np.float32)
        #index: variable that go through halos(can be several galaxies per halo)
        #i: variable that go through galaxies
        #count: find # of galaxies that lie beyond its host halo virial radius
        random.seed(seed) #this is just to check convergence on w_p(r_p)
        index=0; count=0; i=0 
        while (index<halo_mass.size):

            position=halo_pos[index]  #position of the DM halo
            velocity=halo_vel[index]  #velocity of the DM halo
            radius=halo_radius[index] #radius of the DM halo

            #save the position of the central galaxy
            if model=='standard':
                pos_galaxies[i]=position; vel_galaxies[i]=velocity; i+=1

            #if halo contains satellites, save their positions
            Nsat=N_sat[index]
            if Nsat>0:
                offset=halo_offset[index]; length=halo_len[index]
                idss=sorted_ids[IDs[offset:offset+length]]

                #compute the radius of the particles and keep those with R<Rvir
                pos=DM_pos[idss]; posc=pos-position; vel=DM_vel[idss]

                #this is to populate correctly halos closer to box boundaries
                if np.any((position+radius>BoxSize) + (position-radius<0.0)):
                
                    inside=np.where(posc[:,0]>BoxSize/2.0)[0]
                    posc[inside,0]-=BoxSize
                    inside=np.where(posc[:,0]<-BoxSize/2.0)[0]
                    posc[inside,0]+=BoxSize

                    inside=np.where(posc[:,1]>BoxSize/2.0)[0]
                    posc[inside,1]-=BoxSize
                    inside=np.where(posc[:,1]<-BoxSize/2.0)[0]
                    posc[inside,1]+=BoxSize

                    inside=np.where(posc[:,2]>BoxSize/2.0)[0]
                    posc[inside,2]-=BoxSize
                    inside=np.where(posc[:,2]<-BoxSize/2.0)[0]
                    posc[inside,2]+=BoxSize
                
                radii=np.sqrt(posc[:,0]**2+posc[:,1]**2+posc[:,2]**2)
                inside=np.where(radii<radius)[0]
                if len(inside)<Nsat:
                    problematic_cases+=1
                    print('problematic case',len(inside),Nsat)
                else:
                    selected=random.sample(inside,Nsat)
                    pos=pos[selected]; vel=vel[selected]

                #aditional, not esential check. Can be comment out
                #posc=pos-position
                #if np.any((posc>BoxSize/2.0) + (posc<-BoxSize/2.0)):
                #    inside=np.where(posc[:,0]>BoxSize/2.0)[0]
                #    posc[inside,0]-=BoxSize
                #    inside=np.where(posc[:,0]<-BoxSize/2.0)[0]
                #    posc[inside,0]+=BoxSize

                #    inside=np.where(posc[:,1]>BoxSize/2.0)[0]
                #    posc[inside,1]-=BoxSize
                #    inside=np.where(posc[:,1]<-BoxSize/2.0)[0]
                #    posc[inside,1]+=BoxSize

                #    inside=np.where(posc[:,2]>BoxSize/2.0)[0]
                #    posc[inside,2]-=BoxSize
                #    inside=np.where(posc[:,2]<-BoxSize/2.0)[0]
                #    posc[inside,2]+=BoxSize
                #r_max=np.max(np.sqrt(posc[:,0]**2+posc[:,1]**2+posc[:,2]**2))
                #if r_max>radius: #check no particles beyond Rv selected 
                #    print position
                #    print radius
                #    print pos
                #    count+=1

                    for j in range(Nsat):
                        pos_galaxies[i]=pos[j]; vel_galaxies[i]=vel[j]; i+=1
                                              
            index+=1

        if verbose:
            print('done')
        #some final checks
        if i!=N_tot:
            print('some galaxies missing:\nregister %d galaxies out of %d'\
                %(i,N_tot))
        if count>0:
            print('error:',count,'particles beyond the virial radius selected')

        self.pos_galaxies=pos_galaxies
        self.vel_galaxies=vel_galaxies
###############################################################################




##### example of use #####
"""
snapshot_fname='/data1/villa/b500p512nu0.6z99np1024tree/snapdir_017/snap_017'
groups_fname='/home/villa/data1/b500p512nu0.6z99np1024tree'
groups_number=17

### HALO CATALOGUE PARAMETERS ###
mass_criteria='t200'
min_mass=2e12 #Msun/h
max_mass=2e15 #Msun/h

### HOD PARAMETERS ###
fiducial_density=0.00111 #mean number density for galaxies with Mr<-21
M1=8e13
alpha=1.4

pos=hod(snapshot_fname,groups_fname,groups_number,min_mass,max_mass,fiducial_density,M1,alpha,mass_criteria,verbose=True)

print pos
"""



# routines for reading headers and data blocks from Gadget snapshot files
# usage e.g.:
#
# import readsnap as rs
# header = rs.snapshot_header("snap_063.0") # reads snapshot header
# print header.massarr
# mass = rs.read_block("snap_063","MASS",parttype=5) # reads mass for particles of type 5, using block names should work for both format 1 and 2 snapshots
# print "mass for", mass.size, "particles read"
# print mass[0:10]
#
# before using read_block, make sure that the description (and order if using format 1 snapshot files) of the data blocks
# is correct for your configuration of Gadget 
#
# for mutliple file snapshots give e.g. the filename "snap_063" rather than "snap_063.0" to read_block
# for snapshot_header the file number should be included, e.g."snap_063.0", as the headers of the files differ
#
# the returned data block is ordered by particle species even when read from a multiple file snapshot

import numpy as np
import os
import sys
import math
  
# ----- class for snapshot header ----- 

class snapshot_header:
  def __init__(self, filename):

    if os.path.exists(filename):
      curfilename = filename
    elif os.path.exists(filename+".0"):
      curfilename = filename+".0"
    else:
      print("file not found:", filename)
      sys.exit()
      
    self.filename = filename  
    f = open(curfilename,'rb')    
    blocksize = np.fromfile(f,dtype=np.int32,count=1)
    if blocksize[0] == 8:
      swap = 0
      format = 2
    elif blocksize[0] == 256:
      swap = 0
      format = 1  
    else:
      blocksize.byteswap(True)
      if blocksize[0] == 8:
        swap = 1
        format = 2
      elif blocksize[0] == 256:
        swap = 1
        format = 1
      else:
        print("incorrect file format encountered when reading header of", filename)
        sys.exit()
    
    self.format = format
    self.swap = swap
    
    if format==2:
      f.seek(16, os.SEEK_CUR)
    
    self.npart = np.fromfile(f,dtype=np.int32,count=6)
    self.massarr = np.fromfile(f,dtype=np.float64,count=6)
    self.time = (np.fromfile(f,dtype=np.float64,count=1))[0]
    self.redshift = (np.fromfile(f,dtype=np.float64,count=1))[0]
    self.sfr = (np.fromfile(f,dtype=np.int32,count=1))[0]
    self.feedback = (np.fromfile(f,dtype=np.int32,count=1))[0]
    self.nall = np.fromfile(f,dtype=np.uint32,count=6)
    self.cooling = (np.fromfile(f,dtype=np.int32,count=1))[0]
    self.filenum = (np.fromfile(f,dtype=np.int32,count=1))[0]
    self.boxsize = (np.fromfile(f,dtype=np.float64,count=1))[0]
    self.omega_m = (np.fromfile(f,dtype=np.float64,count=1))[0]
    self.omega_l = (np.fromfile(f,dtype=np.float64,count=1))[0]
    self.hubble = (np.fromfile(f,dtype=np.float64,count=1))[0]
    
    if swap:
      self.npart.byteswap(True)
      self.massarr.byteswap(True)
      self.time = self.time.byteswap()
      self.redshift = self.redshift.byteswap()
      self.sfr = self.sfr.byteswap()
      self.feedback = self.feedback.byteswap()
      self.nall.byteswap(True)
      self.cooling = self.cooling.byteswap()
      self.filenum = self.filenum.byteswap()
      self.boxsize = self.boxsize.byteswap()
      self.omega_m = self.omega_m.byteswap()
      self.omega_l = self.omega_l.byteswap()
      self.hubble = self.hubble.byteswap()
     
    f.close()
 
# ----- find offset and size of data block ----- 

def find_block(filename, format, swap, block, block_num, only_list_blocks=False):
  if (not os.path.exists(filename)):
      print("file not found:", filename)
      sys.exit()
            
  f = open(filename,'rb')
  f.seek(0, os.SEEK_END)
  filesize = f.tell()
  f.seek(0, os.SEEK_SET)
  
  found = False
  curblock_num = 1
  while ((not found) and (f.tell()<filesize)):
    if format==2:
      f.seek(4, os.SEEK_CUR)
      curblock = f.read(4).decode()
      if (block == curblock):
        found = True
      f.seek(8, os.SEEK_CUR)  
    else:
      if curblock_num==block_num:
        found = True
        
    curblocksize = (np.fromfile(f,dtype=np.uint32,count=1))[0]
    if swap:
      curblocksize = curblocksize.byteswap()
    
    # - print some debug info about found data blocks -
    #if format==2:
    #  print curblock, curblock_num, curblocksize
    #else:
    #  print curblock_num, curblocksize
    
    if only_list_blocks:
      if format==2:
        print(curblock_num,curblock,f.tell(),curblocksize)
      else:
        print(curblock_num,f.tell(),curblocksize)
      found = False
        
    
    if found:
      blocksize = curblocksize
      offset = f.tell()
    else:
      f.seek(curblocksize, os.SEEK_CUR)
      blocksize_check = (np.fromfile(f,dtype=np.uint32,count=1))[0]
      if swap: blocksize_check = blocksize_check.byteswap()
      if (curblocksize != blocksize_check):
        print("something wrong")
        sys.exit()
      curblock_num += 1
  f.close()
      
  if ((not found) and (not only_list_blocks)):
    print("Error: block not found")
    sys.exit()
    
  if (not only_list_blocks):
    return offset,blocksize
 
# ----- read data block -----
#for snapshots with very very large number of particles set nall manually
#for instance nall=np.array([0,2048**3,0,0,0,0]) 
def read_block(filename, block, parttype=-1, physical_velocities=True, 
    arepo=0, no_masses=False, verbose=False, nall=[0,0,0,0,0,0]):
  
  if (verbose):  print("reading block", block)
  
  blockadd=0
  blocksub=0
  
  if arepo==0:
    if (verbose):  print("Gadget format")
    blockadd=0
  if arepo==1:
    if (verbose):  print("Arepo format")
    blockadd=1  
  if arepo==2:
    if (verbose):  print("Arepo extended format")
    blockadd=4  
  if no_masses==True:
    if (verbose):  print("No mass block present")   
    blocksub=1
         
  if parttype not in [-1,0,1,2,3,4,5]:
    print("wrong parttype given");  sys.exit()
  
  if os.path.exists(filename):        
    curfilename = filename
    single_file = True
  elif os.path.exists(filename+".0"):   
    curfilename = filename+".0"
    single_file = False
  else:
    print("file not found:", filename)
    print("and:", curfilename);  sys.exit()
  
  head = snapshot_header(curfilename)
  format   = head.format
  swap     = head.swap
  npart    = head.npart
  massarr  = head.massarr
  filenum  = head.filenum
  redshift = head.redshift
  time     = head.time
  if np.all(nall==[0,0,0,0,0,0]):
    nall = head.nall
  if verbose:  print("FORMAT=", format)
  del head
  
  # - description of data blocks -
  # add or change blocks as needed for your Gadget version
  data_for_type = np.zeros(6,bool) # should be set to "True" below for the species for which data is stored in the data block #by doing this, the default value is False data_for_type=[False,False,False,False,False,False]
  dt = np.float32 # data type of the data in the block
  if block=="POS ":
    data_for_type[:] = True
    dt = np.dtype((np.float32,3))
    block_num = 2
  elif block=="VEL ":
    data_for_type[:] = True
    dt = np.dtype((np.float32,3))
    block_num = 3
  elif block=="ID  ":
    data_for_type[:] = True
    dt = np.uint32
    block_num = 4
#only used for format I, when file structure is HEAD,POS,VEL,ID,ACCE
  elif block=="ACCE":              #This is only for the PIETRONI project
    data_for_type[:] = True        #This is only for the PIETRONI project
    dt = np.dtype((np.float32,3))  #This is only for the PIETRONI project
    block_num = 5                  #This is only for the PIETRONI project
  elif block=="MASS":
    data_for_type[np.where(massarr==0)] = True
    block_num = 5
    if parttype>=0 and massarr[parttype]>0:   
        if (verbose):    print("filling masses according to massarr")
        if single_file:
          return np.ones(npart[parttype],dtype=dt)*massarr[parttype]
        else:
          return np.ones(nall[parttype],dtype=dt)*massarr[parttype]
  elif block=="U   ":
    data_for_type[0] = True
    block_num = 6-blocksub
  elif block=="RHO ":
    data_for_type[0] = True
    block_num = 7-blocksub
  elif block=="VOL ":
    data_for_type[0] = True
    block_num = 8-blocksub 
  elif block=="CMCE":
    data_for_type[0] = True
    dt = np.dtype((np.float32,3))
    block_num = 9-blocksub 
  elif block=="AREA":
    data_for_type[0] = True
    block_num = 10-blocksub
  elif block=="NFAC":
    data_for_type[0] = True
    dt = np.dtype(np.int64)        #depends on code version, most recent hast int32, old MyIDType   
    block_num = 11-blocksub
  elif block=="NE  ":
    data_for_type[0] = True
    block_num = 8+blockadd-blocksub
  elif block=="NH  ":
    data_for_type[0] = True
    block_num = 9+blockadd-blocksub
  elif block=="HSML":
    data_for_type[0] = True
    block_num = 10+blockadd-blocksub
  elif block=="SFR ":
    data_for_type[0] = True
    block_num = 11+blockadd-blocksub
  elif block=="MHI ":                  #This is only for the bias_HI project
    data_for_type[0] = True            #This is only for the bias_HI project
    block_num = 12+blockadd-blocksub   #This is only for the bias_HI project
  elif block=="TEMP":                  #This is only for the bias_HI project
    data_for_type[0] = True            #This is only for the bias_HI project
    block_num = 13+blockadd-blocksub   #This is only for the bias_HI project
  elif block=="AGE ":
    data_for_type[4] = True
    block_num = 12+blockadd-blocksub
  elif block=="Z   ":
    data_for_type[0] = True
    data_for_type[4] = True
    block_num = 13+blockadd-blocksub
  elif block=="BHMA":
    data_for_type[5] = True
    block_num = 14+blockadd-blocksub
  elif block=="BHMD":
    data_for_type[5] = True
    block_num = 15+blockadd-blocksub
  else:
    print("Sorry! Block type", block, "not known!")
    sys.exit()
  # - end of block description -
  
  actual_data_for_type = np.copy(data_for_type)  
  if parttype >= 0:
    actual_data_for_type[:] = False
    actual_data_for_type[parttype] = True
    if data_for_type[parttype]==False:
      print("Error: no data for specified particle type", parttype, "in the block", block)
      sys.exit()
  elif block=="MASS":
    actual_data_for_type[:] = True  
    
  allpartnum = np.int64(0)
  species_offset = np.zeros(6,np.int64)
  for j in range(6):
    species_offset[j] = allpartnum
    if actual_data_for_type[j]:
        if single_file:
            allpartnum += npart[j]
        else:
          allpartnum += nall[j]

  # define the array containing the information
  data = np.empty(allpartnum,dt)

  # loop over all subfiles    
  for i in range(filenum):

    if not(single_file) and filenum>1: 
        curfilename = filename+"."+str(i)
      
    if i>0:
      head  = snapshot_header(curfilename)
      npart = head.npart  
      del head
      
    curpartnum = np.int32(0)
    cur_species_offset = np.zeros(6,np.int64)
    for j in range(6):
      cur_species_offset[j] = curpartnum
      if data_for_type[j]:
        curpartnum += npart[j]
    
    if parttype>=0:
      actual_curpartnum = npart[parttype]      
      add_offset = cur_species_offset[parttype] 
    else:
      actual_curpartnum = curpartnum
      add_offset = np.int32(0)
      
    offset,blocksize = find_block(curfilename,format,swap,block,block_num)
    
    if i==0: # fix data type for ID if long IDs are used
      if block=="ID  ":
        if blocksize == np.dtype(dt).itemsize*curpartnum * 2:
          dt = np.uint64 
        
    if np.dtype(dt).itemsize*curpartnum != blocksize:
      print("something wrong with blocksize! expected =",np.dtype(dt).itemsize*curpartnum,"actual =",blocksize)
      sys.exit()
    
    f = open(curfilename,'rb')
    f.seek(offset + add_offset*np.dtype(dt).itemsize, os.SEEK_CUR)  
    curdat = np.fromfile(f,dtype=dt,count=actual_curpartnum) # read data
    f.close()  
    if swap:  curdat.byteswap(True)  
          
    for j in range(6):
      if actual_data_for_type[j]:
        if block=="MASS" and massarr[j]>0: # add mass block for particles for which the mass is specified in the snapshot header
          data[species_offset[j]:species_offset[j]+npart[j]] = massarr[j]
        else:
          if parttype>=0:
            data[species_offset[j]:species_offset[j]+npart[j]] = curdat
          else:
            data[species_offset[j]:species_offset[j]+npart[j]] = curdat[cur_species_offset[j]:cur_species_offset[j]+npart[j]]
        species_offset[j] += npart[j]

    del curdat

    if single_file:  break


  if physical_velocities and block=="VEL " and redshift!=0:
    data *= math.sqrt(time)

  return data
  
# ----- list all data blocks in a format 2 snapshot file -----

def list_format2_blocks(filename):
  if   os.path.exists(filename):        curfilename = filename
  elif os.path.exists(filename+".0"):   curfilename = filename+".0"
  else:
    print("file not found:", filename);  sys.exit()
  
  head   = snapshot_header(curfilename)
  format = head.format
  swap   = head.swap;  del head
  
  print('GADGET FORMAT ',format)
  if (format != 2):  print("#   OFFSET   SIZE")
  else:              print("#   BLOCK   OFFSET   SIZE")
  print("-------------------------")
  
  find_block(curfilename, format, swap, "XXXX", 0, only_list_blocks=True)
  
  print("-------------------------")

def read_gadget_header(filename):
  if   os.path.exists(filename):      curfilename = filename
  elif os.path.exists(filename+".0"): curfilename = filename+".0"
  else:
    print("file not found:", filename);  sys.exit()

  head=snapshot_header(curfilename)
  print('npar=',head.npart)
  print('nall=',head.nall)
  print('a=',head.time)
  print('z=',head.redshift)
  print('masses=',head.massarr*1e10,'Msun/h')
  print('boxsize=',head.boxsize,'kpc/h')
  print('filenum=',head.filenum)
  print('cooling=',head.cooling)
  print('Omega_m,Omega_l=',head.omega_m,head.omega_l)
  print('h=',head.hubble,'\n')
  
  rhocrit=2.77536627e11 #h**2 M_sun/Mpc**3
  rhocrit=rhocrit/1e9 #h**2M_sun/kpc**3
  
  Omega_CDM=head.nall[1]*head.massarr[1]*1e10/(head.boxsize**3*rhocrit)
  print('DM mass=%.5e  Omega_DM = %.5f'\
    %(head.massarr[1]*1e10, Omega_CDM))
  if head.nall[2]>0 and head.massarr[2]>0:
    Omega_NU=head.nall[2]*head.massarr[2]*1e10/(head.boxsize**3*rhocrit)
    print('NU mass=%.5e  Omega_NU = %.5f'\
        %(head.massarr[2]*1e10, Omega_NU))
    print('Sum of neutrino masses=%.5f eV'\
        %(Omega_NU*head.hubble**2*94.1745))



import numpy as np
import CAMB_library as CL

################################## INPUT ######################################
# neutrino parameters
hierarchy = 'degenerate' #'degenerate', 'normal', 'inverted'
Mnu       = 0.00         #eV
Nnu       = 3            #number of massive neutrinos
Neff      = 3.046

# cosmological parameters
h       = 0.6711
Omega_m = 0.3175
Omega_b = 0.049
Omega_k = 0.0
s8      = 0.834
tau     = None

# initial P(k) parameters
ns           = 0.9624
As           = 2.13e-9
pivot_scalar = 0.05
pivot_tensor = 0.05

# redshifts and k-range
redshifts    = [0.0, 0.5, 1, 2, 3, 127] 
kmax         = 10.0
k_per_logint = 10

# CAMB accuracy
AccuracyBoost  = 5.0 
lSampleBoost   = 5.0 
lAccuracyBoost = 5.0
###############################################################################

result = CL.PkL(Omega_m=Omega_m, Omega_b=Omega_b, h=h, ns=ns, s8=s8,
                Mnu=Mnu, As=As, Omega_k=Omega_k, 
                pivot_scalar=pivot_scalar, pivot_tensor=pivot_tensor, 
                Nnu=Nnu, hierarchy=hierarchy, Neff=Neff, tau=tau,
                redshifts=redshifts, kmax=kmax, k_per_logint=k_per_logint,
                AccuracyBoost=AccuracyBoost, lSampleBoost=lSampleBoost,
                lAccuracyBoost=lAccuracyBoost, verbose=False)

zs  = result.z
k   = result.k
Pkm = result.Pkmm
Tk  = result.Tk

for i,z in enumerate(zs):
    np.savetxt('Pk_m_z=%.1f.txt'%z, np.transpose([k, Pkm[i]]))
    np.savetxt('Tk_z=%.1f.txt'%z, np.transpose(Tk[:,:,i]))




import numpy as np
import sys,os,time
cimport numpy as np
cimport cython
from cython.parallel import prange,parallel
from libc.math cimport sqrt,pow,sin,cos,log,log10,fabs,round,exp
from libc.stdlib cimport malloc, free
from cython.parallel import prange,parallel
import Pk_library as PKL


DEF PI=3.141592653589793

############################################################################
# This routine places the filter on a grid and returns its FT
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef FT_filter(float BoxSize, float R, int dims, Filter, int threads, 
                float kmin=0, float kmax=0):

    cdef int i,j,l,i1,j1,l1,middle,d2
    cdef float R2, R_grid, factor, k, kF
    cdef double normalization
    cdef float[:,:,::1] field
    cdef np.complex64_t[:,:,::1] field_k

    if Filter not in ['Top-Hat','Gaussian', 'Top-Hat-k']:
        raise Exception('Filter %s not implemented!'%Filter)

    middle        = dims//2
    normalization = 0.0
    R_grid        = (R*dims/BoxSize)
    R2            = R_grid**2
    kF            = 2.0*np.pi/BoxSize

    ###### Top-Hat ######
    if Filter=='Top-Hat':
        field = np.zeros((dims,dims,dims), dtype=np.float32)
        for i in prange(dims, nogil=True):
            i1 = i
            if i1>middle:  i1 = i1 - dims

            for j in range(dims):
                j1 = j
                if j1>middle:  j1 = j1 - dims

                for l in range(dims):
                    l1 = l
                    if l1>middle:  l1 = l1 - dims

                    d2 = i1*i1 + j1*j1 + l1*l1
                    if d2<=R2:  
                        field[i,j,l] = 1.0
                        normalization += 1.0

    ###### Gaussian ######
    if Filter=='Gaussian':
        field = np.zeros((dims,dims,dims), dtype=np.float32)
        for i in prange(dims, nogil=True):
            i1 = i
            if i1>middle:  i1 = i1 - dims

            for j in range(dims):
                j1 = j
                if j1>middle:  j1 = j1 - dims

                for l in range(dims):
                    l1 = l
                    if l1>middle:  l1 = l1 - dims

                    d2 = i1*i1 + j1*j1 + l1*l1
                
                    factor = exp(-d2/(2.0*R2))
                    field[i,j,l] = factor
                    normalization += factor

    ##### Top-Hat k #####
    if Filter=='Top-Hat-k':
        field_k = np.zeros((dims,dims,middle+1), dtype=np.complex64)
        for i in prange(dims, nogil=True):
            i1 = i
            if i1>middle:  i1 = i1 - dims

            for j in range(dims):
                j1 = j
                if j1>middle:  j1 = j1 - dims

                for l in range(middle+1):
                    l1 = l
                    if l1>middle:  l1 = l1 - dims

                    # get the wavenumber of the considered mode
                    k = kF*sqrt(i1*i1 + j1*j1 + l1*l1)

                    # just put to 0 the modes outside kmin < k < kmax
                    if k>=kmin and k<kmax:  field_k[i,j,l] = 1.0
                    else:                   field_k[i,j,l] = 0.0

        # dont mess with DC mode
        field_k[0,0,0] = 1.0

        # now to determine the normalization of this filter, go to real-space
        field = PKL.IFFT3Dr_f(np.asarray(field_k), threads)
        normalization = np.sum(field, dtype=np.float64)

    # normalize the field
    for i in prange(dims, nogil=True):
        for j in range(dims):
            for l in range(dims):
                field[i,j,l] = field[i,j,l]/normalization
    
    # FT the field
    field_k = PKL.FFT3Dr_f(np.asarray(field), threads)
    return np.asarray(field_k)

############################################################################
# This routine places the filter on a 2D grid and returns its FT
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef FT_filter_2D(float BoxSize, float R, int grid, Filter, int threads, 
                   float kmin=0, float kmax=0):

    cdef int i,j,i1,j1,middle,d2
    cdef float R2, R_grid, factor, kF, kN, k 
    cdef double normalization
    cdef float[:,::1] field
    cdef np.complex64_t[:,::1] field_k

    if Filter not in ['Top-Hat','Gaussian', 'Top-Hat-k']:
        raise Exception('Filter %s not implemented!'%Filter)

    middle = grid//2
    normalization = 0.0
    R_grid = (R*grid/BoxSize)
    R2 = R_grid**2
    kF = 2.0*np.pi/BoxSize          #fundamental frequency
    kN = 2.0*np.pi/BoxSize*(grid/2) #Nyquist frequency

    ###### Top-Hat ######
    if Filter=='Top-Hat':
        field = np.zeros((grid,grid), dtype=np.float32)
        for i in prange(grid, nogil=True):
            i1 = i
            if i1>middle:  i1 = i1 - grid

            for j in range(grid):
                j1 = j
                if j1>middle:  j1 = j1 - grid

                d2 = i1*i1 + j1*j1
                if d2<=R2:  
                    field[i,j] = 1.0
                    normalization += 1.0

    ###### Gaussian ######
    if Filter=='Gaussian':
        field = np.zeros((grid,grid), dtype=np.float32)
        for i in prange(grid, nogil=True):
            i1 = i
            if i1>middle:  i1 = i1 - grid

            for j in range(grid):
                j1 = j
                if j1>middle:  j1 = j1 - grid

                d2 = i1*i1 + j1*j1
                
                factor = exp(-d2/(2.0*R2))
                field[i,j] = factor
                normalization += factor

    ##### Top-Hat k #####
    if Filter=='Top-Hat-k':
        field_k = np.zeros((grid,middle+1), dtype=np.complex64)
        for i in prange(grid, nogil=True):
            i1 = i
            if i1>middle:  i1 = i1 - grid

            for j in range(middle+1):
                j1 = j
                if j1>middle:  j1 = j1 - grid

                # get the wavenumber of the considered mode
                k = kF*sqrt(i1*i1 + j1*j1)

                # just put to 0 the modes outside kmin < k < kmax
                if k>=kmin and k<kmax:  field_k[i,j].real = 1.0
                else:                   field_k[i,j].real = 0.0

        # dont mess with DC mode
        field_k[0,0].real = 1.0
                
        # now to determine the normalization of this filter, go to real-space
        field = PKL.IFFT2Dr_f(np.asarray(field_k), threads)
        normalization = np.sum(field, dtype=np.float64)

    # normalize the field
    for i in prange(grid, nogil=True):
        for j in range(grid):
            field[i,j] = field[i,j]/normalization
    
    # FT the field
    field_k = PKL.FFT2Dr_f(np.asarray(field), threads)
    return np.asarray(field_k)

############################################################################
# This routine smooths a field with a given filter and returns the smoothed 
# field. Inputs are the field, the FT of the filter and the number of threads
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef field_smoothing(field, np.complex64_t[:,:,::1] filter_k, int threads):

    cdef int i,j,k,dims,middle
    cdef np.complex64_t[:,:,::1] field_k

    # check that dimensions are the same
    dims = field.shape[0];  middle = dims//2
    if field.shape[0]!=filter_k.shape[0]:
        raise Exception('field and filter have different grids!!!')

    ## compute FFT of the field (change this for double precision) ##
    field_k = PKL.FFT3Dr_f(field,threads) 

    # do a loop over the independent modes.
    for i in prange(dims, nogil=True):
        for j in range(dims):
            for k in range(middle+1): #k=[0,1,..,middle] --> kz>0
                field_k[i,j,k] = field_k[i,j,k]*filter_k[i,j,k]
                                       
    # Fourier transform back
    return PKL.IFFT3Dr_f(field_k,threads)

############################################################################
# This routine smooths a 2D field with a given filter and returns the smoothed 
# field. Inputs are the 2D field, the FT of the filter and the number of threads
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef field_smoothing_2D(field, np.complex64_t[:,::1] filter_k, int threads):

    cdef int i,j,grid,middle
    cdef np.complex64_t[:,::1] field_k

    # check that dimensions are the same
    grid = field.shape[0];  middle = grid//2
    if field.shape[0]!=filter_k.shape[0]:
        raise Exception('field and filter have different grids!!!')

    ## compute FFT of the field (change this for double precision) ##
    field_k = PKL.FFT2Dr_f(field,threads) 

    # do a loop over the independent modes.
    for i in prange(grid, nogil=True):
        for j in range(middle+1): #k=[0,1,..,middle] --> kz>0
            field_k[i,j] = field_k[i,j]*filter_k[i,j]
                                       
    # Fourier transform back
    return PKL.IFFT2Dr_f(field_k,threads)



# This script just imports the different Pylians libraries to see if
# there is any missing dependence or broken link
import bias_library
#import correlation_function_library #weave only works for python 2.X
import cosmology_library
import density_field_library
import HI_clusters_library
#import HI_image_library #requires groupcat
#import HI_library
import HOD_library
#import IM_library  #requires healpy, emcee, corner
import integration_library
import MAS_library
import mass_function_library
import Pk_library
import plotting_library
import readfof
import readsnap
import readgadget
import readsubf
#import readsnap_mpi  #requires mpi4py and barely used
import redshift_space_library
import routines
import smoothing_library
import sorting_library
import units_library
import void_library



import numpy as np
import camb
import sys,os

# This routine computes the linear power spectra using CAMB given the input
# cosmological parameters. To do the rescaling with s8 we always need to generate
# the z=0 linear matter Pk, i.e. in redshifts there always should be 0
# PkL.z -------> redshifts [0, 0.5, 1, 2 ...]
# PkL.k -------> wavenumbers
# PkL.s8 ------> array with the values of sigma8
# PkL.Hz ------> array with the values of Hz
# PkL.Pkmm ----> matrix with matter Pk: Pkmm[1,:] = mm P(k) at z[1]
# PkL.Pkcc ----> matrix with matter Pk: Pkcc[1,:] = cc P(k) at z[1]
# PkL.Pkbb ----> matrix with matter Pk: Pkbb[1,:] = bb P(k) at z[1]
# PkL.Pkcb ----> matrix with matter Pk: Pkcb[1,:] = cb P(k) at z[1]
# PkL.Pknn ----> matrix with matter Pk: Pkcc[1,:] = nu P(k) at z[1]
class PkL:
    def __init__(self, Omega_m=0.3175, Omega_b=0.049, h=0.6711, ns=0.9624, s8=None,
                 Mnu=0.0, As=2.13e-9, Omega_k=0.0, 
                 pivot_scalar=0.05, pivot_tensor=0.05, 
                 Nnu=3, hierarchy='degenerate', Neff=3.046, tau=None,
                 redshifts=[0, 0.5, 1, 2, 3], kmax=10.0, k_per_logint=50,
                 AccuracyBoost=5.0, lSampleBoost=5.0,
                 lAccuracyBoost=5.0, verbose=False):

        Omega_c  = Omega_m - Omega_b - Mnu/(93.14*h**2)
        Omega_cb = Omega_c + Omega_b

        pars = camb.CAMBparams()

        # set accuracy of the calculation
        pars.set_accuracy(AccuracyBoost=AccuracyBoost, lSampleBoost=lSampleBoost, 
                          lAccuracyBoost=lAccuracyBoost, HighAccuracyDefault=True, 
                          DoLateRadTruncation=True)

        # set value of the cosmological parameters
        pars.set_cosmology(H0=h*100.0, ombh2=Omega_b*h**2, omch2=Omega_c*h**2, 
                           mnu=Mnu, omk=Omega_k, 
                           neutrino_hierarchy=hierarchy, 
                           num_massive_neutrinos=Nnu, nnu=Neff,
                           tau=tau)
                   
        # set the value of the primordial power spectrum parameters
        pars.InitPower.set_params(As=As, ns=ns, 
                                  pivot_scalar=pivot_scalar, 
                                  pivot_tensor=pivot_tensor)
    
        # set redshifts, k-range and k-sampling
        pars.set_matter_power(redshifts=redshifts, kmax=kmax, 
                              k_per_logint=k_per_logint)

        # compute results
        results = camb.get_results(pars)

        # get raw matter P(k) and transfer functions with weird k-binning
        #k, zs, Pk = results.get_linear_matter_power_spectrum()
        #Tk        = (results.get_matter_transfer_data()).transfer_data

        # interpolate to get Pmm, Pcc...etc
        k,z,Pmm = results.get_matter_power_spectrum(minkh=2e-5, maxkh=kmax, 
                                                    npoints=500, var1=7, var2=7,
                                                    have_power_spectra=True, 
                                                    params=None)

        k,z,Pcc = results.get_matter_power_spectrum(minkh=2e-5, maxkh=kmax, 
                                                    npoints=500, var1=2, var2=2,
                                                    have_power_spectra=True, 
                                                    params=None)

        k,z,Pbb = results.get_matter_power_spectrum(minkh=2e-5, maxkh=kmax, 
                                                    npoints=500, var1=3, var2=3,
                                                    have_power_spectra=True, 
                                                    params=None)

        k,z,Pcb = results.get_matter_power_spectrum(minkh=2e-5, maxkh=kmax, 
                                                    npoints=500, var1=2, var2=3,
                                                    have_power_spectra=True, 
                                                    params=None)

        Pcb = (Omega_c**2*Pcc + Omega_b**2*Pbb +\
               2.0*Omega_b*Omega_c*Pcb)/Omega_cb**2

        k,z,Pnn = results.get_matter_power_spectrum(minkh=2e-5, maxkh=kmax, 
                                                    npoints=500, var1=6, var2=6,
                                                    have_power_spectra=True, 
                                                    params=None)

        # rescale by sigma_8
        s8_linear = results.get_sigma8()[-1]
        if s8!=None and z[0]!=0.0:
            raise Exception('To rescale by s8 we need to generate the linear Pk at z=0')
        factor = (s8/s8_linear)**2

        # get sigma_8 and Hz in km/s/(kpc/h)
        self.s8 = np.array(results.get_sigma8())[::-1]*np.sqrt(factor)
        self.Hz = np.array([results.hubble_parameter(red) for red in z])

        # get transfer function
        Tk  = (results.get_matter_transfer_data()).transfer_data
        #for j,z in enumerate(redshifts):
        #    fout = 'ICs/Tk_m_z=%s.txt'%z
        #    np.savetxt(fout, np.transpose(Tk[:,:,j]))

        self.z    = z;    self.k   = k 
        self.Pkmm = Pmm*factor;  self.Pknn = Pnn*factor
        self.Pkcc = Pcc*factor;  self.Pkbb = Pbb*factor;  self.Pkcb = Pcb*factor
        self.Tk   = Tk

        if verbose:  print(pars)
        
        #fout = 'Pk_trans_z=%.3f.txt'%z
        # notice that transfer functions have an inverted order:i=0 ==>z_max
        #np.savetxt(fout,np.transpose([Tk[0,:,i],Tk[1,:,i],Tk[2,:,i],Tk[3,:,i],
        #                               Tk[4,:,i],Tk[5,:,i],Tk[6,:,i]]))



import numpy as np 
import time,sys,os
cimport numpy as np
cimport cython
from libc.math cimport sqrt,pow,sin,cos,floor,fabs
cimport MAS_c as MASC
import Pk_library as PKL
import MAS_library as MASL

# This routine computes the gravitational potential of a 3D density field
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef potential(delta, float Omega_m, float a, MAS='CIC', threads=1):

    cdef int kxx, kyy, kzz, kx, ky, kz, grid, middle, MAS_index
    cdef double green_func, prefact
    cdef double MAS_corr[3]
    ####### change this for double precision ######
    cdef float MAS_factor
    cdef np.complex64_t[:,:,::1] delta_k, phi_k
    ###############################################

    # get dimensions of the grid
    grid = delta.shape[0];  middle = grid//2
    MAS_index = PKL.MAS_function(MAS)

    ## compute FFT of the field (change this for double precision) ##
    delta_k = PKL.FFT3Dr_f(delta, threads)

    # define the potential field
    phi_k = np.zeros((grid,grid,middle+1), dtype=np.complex64)

    # do a loop over the independent modes.
    # compute k,k_par,k_per, mu for each mode. k's are in kF units
    start2 = time.time();  prefact = np.pi/grid
    for kxx in range(grid):
        kx = (kxx-grid if (kxx>middle) else kxx)
        MAS_corr[0] = PKL.MAS_correction(prefact*kx,MAS_index)
        
        for kyy in range(grid):
            ky = (kyy-grid if (kyy>middle) else kyy)
            MAS_corr[1] = PKL.MAS_correction(prefact*ky,MAS_index)

            for kzz in range(middle+1): #kzz=[0,1,..,middle] --> kz>0
                kz = (kzz-grid if (kzz>middle) else kzz)
                MAS_corr[2] = PKL.MAS_correction(prefact*kz,MAS_index)  

                # avoid DC mode
                if (kx==0 and ky==0 and kz==0):  continue

                # correct modes amplitude for MAS
                MAS_factor = MAS_corr[0]*MAS_corr[1]*MAS_corr[2]
                delta_k[kxx,kyy,kzz] = delta_k[kxx,kyy,kzz]*MAS_factor

                # compute Green function
                green_func = -3.0*Omega_m/(8.0*a)/(sin(kx*prefact)**2 + sin(ky*prefact)**2 + sin(kz*prefact)**2)

                # multiply by Green function
                phi_k[kxx,kyy,kzz] = delta_k[kxx,kyy,kzz]*green_func

    return PKL.IFFT3Dr_f(phi_k,threads)



# This routine computes the tidal tensor of a 3D density field
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
class tidal_tensor:
    def __init__(self, delta, float Omega_m, float a, MAS='CIC', threads=1):

        cdef int kxx, kyy, kzz, kx, ky, kz, grid, middle, MAS_index
        cdef double green_func, prefact
        cdef float complex phi_k
        cdef double MAS_corr[3]
        ####### change this for double precision ######
        cdef float MAS_factor
        cdef np.complex64_t[:,:,::1] delta_k, T00_k, T01_k, T02_k, T11_k, T12_k, T22_k
        ###############################################

        # get dimensions of the grid
        grid = delta.shape[0];  middle = grid//2
        MAS_index = PKL.MAS_function(MAS)

        # define the different tidal tensor arrays
        T00_k = np.zeros((grid,grid,middle+1), dtype=np.complex64)
        T01_k = np.zeros((grid,grid,middle+1), dtype=np.complex64)
        T02_k = np.zeros((grid,grid,middle+1), dtype=np.complex64)
        T11_k = np.zeros((grid,grid,middle+1), dtype=np.complex64)
        T12_k = np.zeros((grid,grid,middle+1), dtype=np.complex64)
        T22_k = np.zeros((grid,grid,middle+1), dtype=np.complex64)

        ## compute FFT of the field (change this for double precision) ##
        delta_k = PKL.FFT3Dr_f(delta, threads)

        # do a loop over all modes
        prefact = np.pi/grid
        for kxx in range(grid):
            kx = (kxx-grid if (kxx>middle) else kxx)
            MAS_corr[0] = PKL.MAS_correction(prefact*kx,MAS_index)
        
            for kyy in range(grid):
                ky = (kyy-grid if (kyy>middle) else kyy)
                MAS_corr[1] = PKL.MAS_correction(prefact*ky,MAS_index)

                for kzz in range(middle+1): #kzz=[0,1,..,middle] --> kz>0
                    kz = (kzz-grid if (kzz>middle) else kzz)
                    MAS_corr[2] = PKL.MAS_correction(prefact*kz,MAS_index)  

                    # avoid DC mode
                    if (kx==0 and ky==0 and kz==0):  continue

                    # correct modes amplitude for MAS
                    MAS_factor = MAS_corr[0]*MAS_corr[1]*MAS_corr[2]
                    delta_k[kxx,kyy,kzz] = delta_k[kxx,kyy,kzz]*MAS_factor

                    # compute Green function
                    green_func = -3.0*Omega_m/(8.0*a)/(sin(kx*prefact)**2 + sin(ky*prefact)**2 + sin(kz*prefact)**2)

                    # get each component of the tidal tensor
                    phi_k              = delta_k[kxx,kyy,kzz]*green_func
                    T00_k[kxx,kyy,kzz] = phi_k*kx*kx
                    T01_k[kxx,kyy,kzz] = phi_k*kx*ky
                    T02_k[kxx,kyy,kzz] = phi_k*kx*kz
                    T11_k[kxx,kyy,kzz] = phi_k*ky*ky
                    T12_k[kxx,kyy,kzz] = phi_k*ky*kz
                    T22_k[kxx,kyy,kzz] = phi_k*kz*kz

                
        self.T00 = PKL.IFFT3Dr_f(T00_k,threads)
        self.T01 = PKL.IFFT3Dr_f(T01_k,threads)
        self.T02 = PKL.IFFT3Dr_f(T02_k,threads)
        self.T11 = PKL.IFFT3Dr_f(T11_k,threads)
        self.T12 = PKL.IFFT3Dr_f(T12_k,threads)
        self.T22 = PKL.IFFT3Dr_f(T22_k,threads)






# This library is used to make images from the Illustris simulations
from mpi4py import MPI
import numpy as np
#import snapshot as sn
import groupcat
import readsnapHDF5 as rs
import HI_library as HIL
import sys,os,glob,h5py,hdf5plugin,time
import MAS_library as MASL
import redshift_space_library as RSL

####### MPI DEFINITIONS #######
comm   = MPI.COMM_WORLD
nprocs = comm.Get_size()
myrank = comm.Get_rank()

#######################################################################################
# This routine reads in parallel the whole Illustris snapshot and select the 
# particles within the input region
# snapshot_root -------> folder containing the snapdir folders
# snapnum -------------> number of the snapshot
# TREECOOL_file -------> location of TREECOOL file
# x_min, x_max --------> limits in the x-direction
# y_min, y_max --------> limits in the y-direction
# z_min, z_max --------> limits in the z-direction
# padding -------------> padding to get all particles are borders
# fout ----------------> name of output file (hdf5)
# redshift_space ------> whether move particles to redshift-space
# axis ----------------> axis along which move particles in redshift-space
def Illustris_region(snapshot_root, snapnum, TREECOOL_file, x_min, x_max, 
                     y_min, y_max, z_min, z_max, padding, fout,
                     redshift_space=False, axis=0):


    # read snapshot and find number of subfiles
    snapshot = snapshot_root + 'snapdir_%03d/snap_%03d'%(snapnum,snapnum)
    header   = rs.snapshot_header(snapshot)
    nall     = header.nall
    redshift = header.redshift
    BoxSize  = header.boxsize/1e3 #Mpc/h
    filenum  = header.filenum
    Omega_m  = header.omega0
    Omega_L  = header.omegaL
    h        = header.hubble
    Hubble = 100.0*np.sqrt(Omega_m*(1.0+redshift)**3+Omega_L) #km/s/(Mpc/h)

    if myrank==0:
        print('\n')
        print('BoxSize         = %.3f Mpc/h'%BoxSize)
        print('Number of files = %d'%filenum)
        print('Omega_m         = %.3f'%Omega_m)
        print('Omega_l         = %.3f'%Omega_L)
        print('redshift        = %.3f'%redshift)

    # find the numbers each cpu will work on
    array   = np.arange(0, filenum)
    numbers = np.where(array%nprocs==myrank)[0]

    # do a loop over the different realizations
    particles = 0
    for i in numbers:

        snapshot = snapshot_root + 'snapdir_%03d/snap_%03d.%d'%(snapnum,snapnum,i)
        pos = rs.read_block(snapshot, 'POS ', parttype=0, verbose=False)/1e3
        pos = pos.astype(np.float32)

        # read velocities and displace particle positions
        if redshift_space:
            vel = rs.read_block(snapshot, 'VEL ', parttype=0, verbose=False)/np.sqrt(1.0+redshift) #km/s
            RSL.pos_redshift_space(pos, vel, BoxSize, Hubble, redshift, axis)

        # check if particles are in the region
        indexes_region = np.where((pos[:,0]>=x_min-padding) & (pos[:,0]<=x_max+padding) &\
                                  (pos[:,1]>=y_min-padding) & (pos[:,1]<=y_max+padding) &\
                                  (pos[:,2]>=z_min-padding) & (pos[:,2]<=z_max+padding))[0]

        # if particles are not in the region continue
        local_particles = indexes_region.shape[0]
        print('Myrank = %d ---> num = %d ---> part = %ld'%(myrank,i,local_particles))
        if local_particles==0:  continue

        # find radii, HI and gas masses
        MHI  = rs.read_block(snapshot, 'NH  ', parttype=0, verbose=False)#HI/H
        mass = rs.read_block(snapshot, 'MASS', parttype=0, verbose=False)*1e10
        SFR  = rs.read_block(snapshot, 'SFR ', parttype=0, verbose=False)
        indexes = np.where(SFR>0.0)[0];  del SFR

        # find the metallicity of star-forming particles
        metals = rs.read_block(snapshot, 'GZ  ', parttype=0, verbose=False)
        metals = metals[indexes]/0.0127

        # find densities of star-forming particles: units of h^2 Msun/Mpc^3
        rho    = rs.read_block(snapshot, 'RHO ', parttype=0, verbose=False)*1e19
        Volume = mass/rho                            #(Mpc/h)^3
        radii  = (Volume/(4.0*np.pi/3.0))**(1.0/3.0) #Mpc/h 
        rho    = rho[indexes]                        #h^2 Msun/Mpc^3
        Volume = Volume[indexes]                     #(Mpc/h)^3

        # find volume and radius of star-forming particles
        radii_SFR  = (Volume/(4.0*np.pi/3.0))**(1.0/3.0) #Mpc/h 
            
        # find HI/H fraction for star-forming particles
        MHI[indexes] = HIL.Rahmati_HI_Illustris(rho, radii_SFR, metals, redshift, 
                                                h, TREECOOL_file, Gamma=None,
                                                fac=1, correct_H2=True) #HI/H
        MHI *= (0.76*mass)
            

        # select the particles belonging to the region
        pos   = pos[indexes_region]
        MHI   = MHI[indexes_region]
        radii = radii[indexes_region]
        mass  = mass[indexes_region]

        # write partial files        
        new_size = particles + local_particles    

        if particles==0:
            f = h5py.File(fout[:-5]+'_%d.hdf5'%myrank, 'w')
            f.create_dataset('pos',   data=pos,   maxshape=(None,3))
            f.create_dataset('M_HI',  data=MHI,   maxshape=(None,))
            f.create_dataset('radii', data=radii, maxshape=(None,))
            f.create_dataset('mass',  data=mass,  maxshape=(None,))
        else:
            f = h5py.File(fout[:-5]+'_%d.hdf5'%myrank, 'a')
            pos_f   = f['pos'];    pos_f.resize((new_size,3))
            M_HI_f  = f['M_HI'];   M_HI_f.resize((new_size,))
            radii_f = f['radii'];  radii_f.resize((new_size,))
            mass_f  = f['mass'];   mass_f.resize((new_size,))
            pos_f[particles:]   = pos
            M_HI_f[particles:]  = MHI
            radii_f[particles:] = radii
            mass_f[particles:]  = mass
        f.close()
        particles += local_particles
                
    comm.Barrier()

    # sum the particles found in each cpu
    All_particles = 0 
    All_particles = comm.reduce(particles, op=MPI.SUM, root=0)

    # Master will merge partial files into a file one
    if myrank==0:

        print('Found %d particles'%All_particles)
        f = h5py.File(fout,'w')
        
        f1 = h5py.File(fout[:-5]+'_0.hdf5','r')
        pos   = f1['pos'][:]
        M_HI  = f1['M_HI'][:]
        radii = f1['radii'][:]
        mass  = f1['mass'][:]
        f1.close()

        particles = mass.shape[0]
        pos_f   = f.create_dataset('pos',   data=pos,   maxshape=(None,3))
        M_HI_f  = f.create_dataset('M_HI',  data=M_HI,  maxshape=(None,))
        radii_f = f.create_dataset('radii', data=radii, maxshape=(None,))
        mass_f  = f.create_dataset('mass',  data=mass,  maxshape=(None,))

        for i in range(1,nprocs):
            f1 = h5py.File(fout[:-5]+'_%d.hdf5'%i,'r')
            pos   = f1['pos'][:]
            M_HI  = f1['M_HI'][:]
            radii = f1['radii'][:]
            mass  = f1['mass'][:]
            f1.close()
            
            size = mass.shape[0]
            
            pos_f.resize((particles+size,3));  pos_f[particles:] = pos
            M_HI_f.resize((particles+size,));  M_HI_f[particles:] = M_HI
            radii_f.resize((particles+size,)); radii_f[particles:] = radii
            mass_f.resize((particles+size,));  mass_f[particles:] = mass

            particles += size

        f.close()

        for i in range(nprocs):
            os.system('rm '+fout[:-5]+'_%d.hdf5'%i)
#######################################################################################

################################################################################
# This routine finds the positions, radii, HI and gas masses of the particles 
# belonging to a give halo
def Illustris_halo(snapshot_root, snapnum, halo_number, TREECOOL_file, fout,
                   ptype=0):

    # find snapshot name and read header
    snapshot = snapshot_root + 'snapdir_%03d/snap_%03d'%(snapnum, snapnum)
    header   = rs.snapshot_header(snapshot)
    redshift = header.redshift
    BoxSize  = header.boxsize/1e3 #Mpc/h
    filenum  = header.filenum
    Omega_m  = header.omega0
    Omega_L  = header.omegaL
    h        = header.hubble
    massarr  = header.massarr*1e10 #Msun/h

    print('\nBoxSize         = %.1f Mpc/h'%BoxSize)
    print('Number of files = %d'%filenum)
    print('Omega_m         = %.3f'%Omega_m)
    print('Omega_l         = %.3f'%Omega_L)
    print('redshift        = %.3f'%redshift)

    # read number of particles in halos and subhalos and number of subhalos
    halos = groupcat.loadHalos(snapshot_root, snapnum, 
            fields=['GroupLenType','GroupPos','GroupMass'])
    halo_len  = halos['GroupLenType'][:,ptype]  
    halo_pos  = halos['GroupPos']/1e3
    halo_mass = halos['GroupMass']*1e10
    del halos


    # find where the halo starts and ends in the file
    begin = np.sum(halo_len[:halo_number], dtype=np.int64)
    end   = begin + halo_len[halo_number]
    print(begin,end)

    # do a loop over all snapshot subfiles
    f = h5py.File(fout,'w')
    pos_f   = f.create_dataset('pos',   (0,3),  maxshape=(None,3))
    vel_f   = f.create_dataset('vel',   (0,3),  maxshape=(None,3))
    if ptype==0:
        mass_f  = f.create_dataset('mass',  (0,),   maxshape=(None,))
        MHI_f   = f.create_dataset('M_HI',  (0,),   maxshape=(None,))
        radii_f = f.create_dataset('radii', (0,),   maxshape=(None,))
    if ptype==1:
        radii_f = f.create_dataset('radii',  (0,),   maxshape=(None,))
        mass_f  = f.create_dataset('mass_c', (0,),   maxshape=(None,))

    begin_subfile, particles = 0, 0
    for i in range(filenum):

        # find subfile name and read the number of particles in it
        snapshot = snapshot_root + 'snapdir_%03d/snap_%03d.%d'%(snapnum, snapnum, i)
        header = rs.snapshot_header(snapshot)
        npart  = header.npart 

        end_subfile = begin_subfile + npart[ptype]

        # if all particles in the halo has been read exit loop
        if end<begin_subfile:  break

        # if the subfile does not contain any particle move to next subfile
        if begin>end_subfile:
            begin_subfile = end_subfile;  continue


        print('Working with subfile %03d'%i)
        pos  = rs.read_block(snapshot, 'POS ', parttype=ptype, 
                             verbose=False)/1e3
        pos  = pos.astype(np.float32)
        vel  = rs.read_block(snapshot, 'VEL ', parttype=ptype, 
                             verbose=False)/np.sqrt(1.0+redshift) #km/s

        if ptype==0:
            MHI  = rs.read_block(snapshot, 'NH  ', parttype=0,
                                 verbose=False)#HI/H
            mass = rs.read_block(snapshot, 'MASS', parttype=0,
                                 verbose=False)*1e10
            SFR  = rs.read_block(snapshot, 'SFR ', parttype=0,
                                 verbose=False)
            indexes = np.where(SFR>0.0)[0];  del SFR

            # find the metallicity of star-forming particles
            metals = rs.read_block(snapshot, 'GZ  ', parttype=0, verbose=False)
            metals = metals[indexes]/0.0127

            # find densities of star-forming particles: units of h^2 Msun/Mpc^3
            rho = rs.read_block(snapshot, 'RHO ', parttype=0, 
                                verbose=False)*1e19
            Volume = mass/rho                            #(Mpc/h)^3
            radii  = (Volume/(4.0*np.pi/3.0))**(1.0/3.0) #Mpc/h 

            # find density and radius of star-forming particles
            radii_SFR = radii[indexes]    
            rho       = rho[indexes]

            # find HI/H fraction for star-forming particles
            MHI[indexes] = HIL.Rahmati_HI_Illustris(rho, radii_SFR, metals,
                                                    redshift, h, TREECOOL_file,
                                                    Gamma=None, fac=1,
                                                    correct_H2=True) #HI/H
            MHI *= (0.76*mass)
            
        if ptype==1:
            radii = rs.read_block(snapshot, 'SFHS', parttype=1,
                                  verbose=False)/1e3 #Mpc/h
            mass = np.ones(len(radii))*massarr[1]
            

        # find the indexes of current subfile that contribute to halo
        begin_array = begin - begin_subfile
        end_array   = begin_array + (end-begin)

        if end>end_subfile:
            end_array = end_subfile - begin_subfile
            begin     = end_subfile

        new_size = particles + (end_array - begin_array)

        pos_f.resize((new_size,3))
        pos_f[particles:] = pos[begin_array:end_array]
        vel_f.resize((new_size,3))
        vel_f[particles:] = vel[begin_array:end_array]

        if ptype==0:
            mass_f.resize((new_size,))
            mass_f[particles:] = mass[begin_array:end_array]

            MHI_f.resize((new_size,))
            MHI_f[particles:] = MHI[begin_array:end_array]

            radii_f.resize((new_size,))
            radii_f[particles:] = radii[begin_array:end_array]

        if ptype==1:
            radii_f.resize((new_size,))
            radii_f[particles:] = radii[begin_array:end_array]

            mass_f.resize((new_size,))
            mass_f[particles:] = mass[begin_array:end_array]

        particles = new_size
        begin_subfile = end_subfile


    f.close()
    print('Halo mass = %.3e'%halo_mass[halo_number])
    print('Halo pos  =',halo_pos[halo_number])
    print('Number of particles in the halo = %ld'%particles)




################################################################################
# This class takes as input the halo number and returns the positions, 
# velocities and HI masses of the particles in it. Can be for gas, CDM...etc
class particles_in_halo:
    def __init__(self, snapshot_root, snapnum, halo_number, TREECOOL_file,
                 ptype):

        # read header
        snapshot = snapshot_root + 'snapdir_%03d/snap_%03d'%(snapnum, snapnum)
        f            = h5py.File(snapshot+'.0.hdf5', 'r')
        scale_factor = f['Header'].attrs[u'Time']
        redshift     = f['Header'].attrs[u'Redshift']
        BoxSize      = f['Header'].attrs[u'BoxSize']/1e3  #Mpc/h    
        filenum      = f['Header'].attrs[u'NumFilesPerSnapshot']
        Omega_m      = f['Header'].attrs[u'Omega0']
        Omega_L      = f['Header'].attrs[u'OmegaLambda']
        h            = f['Header'].attrs[u'HubbleParam']
        Masses       = f['Header'].attrs[u'MassTable']*1e10  #Msun/h     
        f.close()

        print('\nBoxSize         = %.1f Mpc/h'%BoxSize)
        print('Number of files = %d'%filenum)
        print('Omega_m         = %.3f'%Omega_m)
        print('Omega_l         = %.3f'%Omega_L)
        print('redshift        = %.3f'%redshift)

        # read number of particles in halos and subhalos and number of subhalos
        halos = groupcat.loadHalos(snapshot_root, snapnum, 
                                   fields=['GroupLenType','GroupPos',
                                           'GroupMass'])
        halo_len  = halos['GroupLenType'][:,ptype]  
        halo_pos  = halos['GroupPos']/1e3
        halo_mass = halos['GroupMass']*1e10
        del halos

        # find where the halo starts and ends in the file
        begin = np.sum(halo_len[:halo_number], dtype=np.int64)
        end   = begin + halo_len[halo_number]
        length = end-begin
        print(begin,end)

        self.pos = np.zeros((length,3), dtype=np.float32)
        self.vel = np.zeros((length,3), dtype=np.float32)
        
        if ptype==0:
            self.M_HI = np.zeros(length, dtype=np.float32)


        begin_subfile, particles = 0, 0
        for i in range(filenum):

            # find subfile name and read the number of particles in it
            snapshot = snapshot_root + 'snapdir_%03d/snap_%03d.%d.hdf5'%(snapnum, snapnum, i)
            f     = h5py.File(snapshot, 'r')
            npart = f['Header'].attrs[u'NumPart_ThisFile']
            f.close()

            end_subfile = begin_subfile + npart[ptype]

            # if all particles in the halo has been read exit loop
            if end<begin_subfile:  break

            # if the subfile does not contain any particle move to next subfile
            if begin>end_subfile:
                begin_subfile = end_subfile;  continue

            print(i)
            f     = h5py.File(snapshot, 'r')
            print('Working with subfile %03d'%i)
            pos_i = (f['PartType%d/Coordinates'%ptype][:]/1e3).astype(np.float32)
            vel_i = f['PartType%d/Velocities'%ptype][:]*np.sqrt(scale_factor)

            if ptype==0:
                pos_i, M_HI_i = HIL.HI_mass_from_Illustris_snap(snapshot, TREECOOL_file)

            # find the indexes of current subfile that contribute to halo
            begin_array = begin - begin_subfile
            end_array   = begin_array + length

            if end>end_subfile:
                end_array = end_subfile - begin_subfile
                begin     = end_subfile

            length_array = end_array-begin_array
                
            self.pos[particles:particles+length_array]  = pos_i[begin_array:end_array]
            self.vel[particles:particles+length_array]  = vel_i[begin_array:end_array]
            if ptype==0:
                self.M_HI[particles:particles+length_array] = M_HI_i[begin_array:end_array]

            particles = particles+length_array
            begin_subfile = end_subfile
            f.close()

        print('Halo mass = %.3e'%halo_mass[halo_number])
        print('Halo pos  =',halo_pos[halo_number])
        print('Number of particles in the halo = %ld'%particles)



import numpy as np 
import time,sys,os
import readsnap,readgadget
import redshift_space_library as RSL
import MAS_library as MASL

def density_field_gadget(snapshot_fname, ptypes, dims, MAS='CIC',
                         do_RSD=False, axis=0, verbose=True): 

    start = time.time()
    if verbose:  print('\nComputing density field of particles',ptypes)

    # declare the array hosting the density field
    density = np.zeros((dims, dims, dims), dtype=np.float32)

    # read relevant paramaters on the snapshot
    head     = readgadget.header(snapshot_fname)
    BoxSize  = head.boxsize/1e3 #Mpc/h
    Masses   = head.massarr*1e10 #Msun/h
    Nall     = head.nall;  Ntotal = np.sum(Nall,dtype=np.int64)
    filenum  = head.filenum
    Omega_m  = head.omega_m
    Omega_l  = head.omega_l
    redshift = head.redshift
    fformat  = head.format
    Hubble   = head.Hubble


    if ptypes==[-1]:    ptypes = [0, 1, 2, 3, 4, 5]
    if len(ptypes)==1:  single_component=True
    else:               single_component=False

    # do a loop over all files
    num = 0.0
    for i in range(filenum):

        # find the name of the sub-snapshot
        if filenum==1:       snapshot = snapshot_fname
        else:                snapshot = snapshot_fname+'.%d'%i
        if fformat=='hdf5':  snapshot = snapshot+'.hdf5'

        # find the local particles in the sub-snapshot
        head  = readgadget.header(snapshot)
        npart = head.npart

        # do a loop over all particle types
        for ptype in ptypes:

            if npart[ptype]==0:  continue

            # read positions in Mpc/h
            pos = readgadget.read_field(snapshot, "POS ", ptype)/1e3
            #pos = readsnap.read_block(snapshot,"POS ",parttype=ptype)/1e3

            # read velocities in km/s and move particles to redshift-space
            if do_RSD:
                vel = readgadget.read_field(snapshot, "VEL ", ptype)
                #vel = readsnap.read_block(snapshot,"VEL ",parttype=ptype)
                RSL.pos_redshift_space(pos,vel,BoxSize,Hubble,redshift,axis)
                del vel

            # compute density field. If multicomponent, read/find masses
            if Masses[ptype]!=0:
                if single_component:
                    MASL.MA(pos, density, BoxSize, MAS) 
                    num += pos.shape[0]
                else:
                    mass = np.ones(npart[ptype], dtype=np.float32)*Masses[ptype]
                    MASL.MA(pos, density, BoxSize, MAS, W=mass)
                    num += np.sum(mass, dtype=np.float64)
            else:
                mass = readgadget.read_field(snapshot, "MASS", ptype)*1e10
                #mass = readsnap.read_block(snapshot,"MASS",
                #        parttype=ptype)*1e10 #Msun/h
                MASL.MA(pos, density, BoxSize, MAS, W=mass)
                num += np.sum(mass, dtype=np.float64)

    if verbose:
        print('%.8e should be equal to\n%.8e'\
            %(np.sum(density, dtype=np.float64), num))
        print('Time taken = %.2f seconds'%(time.time()-start))

    return np.asarray(density)


















# This library contains scripts that can be used to plot density fields
import numpy as np
import readgadget
import MAS_library as MASL
import sys,os

# This routine finds the name of the density field
def density_field_name(snapshot, x_min, x_max, y_min, y_max, 
                       z_min, z_max, dims, ptypes, plane, MAS):

    # this is a number that describes the ptypes used.
    # ptypes = [1] -----> part_num = 2
    # ptypes = [0,1] ---> part_num = 3
    # ptypes = [0,4] ---> part_num = 17
    part_num = np.sum(2**np.array(ptypes))

    # name of the density field
    f_df = 'density_field_%.3f_%.3f_%.3f_%.3f_%.3f_%.3f_%d_%d_%s_%s_%s.npy'\
           %(x_min, x_max, y_min, y_max, z_min, z_max, dims, part_num, 
             plane, MAS, snapshot[-3:])

    return f_df

# This routine computes the coordinates of the density field square 
def geometry(snapshot, plane, x_min, x_max, y_min, y_max, z_min, z_max):

    # read snapshot head and obtain BoxSize
    head    = readgadget.header(snapshot)
    BoxSize = head.boxsize/1e3 #Mpc/h                    

    plane_dict = {'XY':[0,1], 'XZ':[0,2], 'YZ':[1,2]}

    # check that the plane is square
    if plane=='XY':
        length1 = x_max-x_min;  length2 = y_max-y_min;  depth = z_max-z_min 
        offset1 = x_min;        offset2 = y_min
    elif plane=='XZ':
        length1 = x_max-x_min;  length2 = z_max-z_min;  depth = y_max-y_min 
        offset1 = x_min;        offset2 = z_min
    else:
        length1 = y_max-y_min;  length2 = z_max-z_min;  depth = x_max-x_min 
        offset1 = y_min;        offset2 = z_min
    if length1!=length2:
        raise Exception('Plane has to be a square!!')
    BoxSize_slice = length1

    return length1, offset1, length2, offset2, depth, BoxSize_slice


# This routine reads the positions (and masses) of the particle type choosen
# and computes the 2D density field using NGP, CIC, TSC or PCS
def density_field_2D(snapshot, x_min, x_max, y_min, y_max, z_min, z_max,
                     dims, ptypes, plane, MAS, save_density_field):


    # find the geometric values of the density field square
    dx, x, dy, y, depth, BoxSize_slice = \
        geometry(snapshot, plane, x_min, x_max, y_min, y_max, z_min, z_max)

    # find the name of the density field and read it if already exists
    f_df = density_field_name(snapshot, x_min, x_max, y_min, y_max, 
                                 z_min, z_max, dims, ptypes, plane, MAS)
    if os.path.exists(f_df):
        print('\nDensity field already computed. Reading it from file...')
        overdensity = np.load(f_df);  return dx, x, dy, y, overdensity

    # if not, compute it
    print('\nComputing density field...')
    plane_dict = {'XY':[0,1], 'XZ':[0,2], 'YZ':[1,2]}

    # read snapshot head and obtain BoxSize, filenum...
    head     = readgadget.header(snapshot)
    BoxSize  = head.boxsize/1e3 #Mpc/h                    
    Nall     = head.nall
    Masses   = head.massarr*1e10 #Msun/h                  
    filenum  = head.filenum
    redshift = head.redshift

    # define the density array
    overdensity = np.zeros((dims,dims), dtype=np.float32)

    # do a loop over all subfiles in the snapshot
    total_mass, mass_slice = 0.0, 0.0;  renormalize_2D = False
    for i in range(filenum):

        # find the name of the subfile
        snap1 = '%s.%d'%(snapshot,i)
        snap2 = '%s.%d.hdf5'%(snapshot,i)
        snap3 = '%s'%(snapshot)
        if   os.path.exists(snap1):  snap = snap1
        elif os.path.exists(snap2):  snap = snap2
        elif os.path.exists(snap3):  snap = snap3
        else:  raise Exception('Problem with the snapshot name!')                

        # in the last subfile we renormalize the field
        if i==filenum-1:  renormalize_2D = True

        # do a loop over all particle types
        for ptype in ptypes:

            # read the positions of the particles in Mpc/h
            pos = readgadget.read_field(snap,"POS ",ptype)/1e3
            
            # keep only with the particles in the slice
            indexes = np.where((pos[:,0]>x_min) & (pos[:,0]<x_max) &
                               (pos[:,1]>y_min) & (pos[:,1]<y_max) &
                               (pos[:,2]>z_min) & (pos[:,2]<z_max) )
            pos = pos[indexes]

            # renormalize positions
            pos[:,0] -= x_min;  pos[:,1] -= y_min;  pos[:,2] -= z_min

            # project particle positions into a 2D plane
            pos = pos[:,plane_dict[plane]]

            # read the masses of the particles
            mass = readgadget.read_field(snap,"MASS",ptype)*1e10
            total_mass += np.sum(mass, dtype=np.float64)
            mass = mass[indexes];  mass_slice += np.sum(mass)

            # update 2D density field
            MASL.MA(pos, overdensity, BoxSize_slice, MAS=MAS, W=mass,
                    renormalize_2D=renormalize_2D)
            

    print('Expected mass = %.7e'%mass_slice)
    print('Computed mass = %.7e'%np.sum(overdensity, dtype=np.float64))

    # compute mean density in the whole box
    mass_density = total_mass*1.0/BoxSize**3 #(Msun/h)/(Mpc/h)^3 or #/(Mpc/h)^3
    print('mass density = %.5e'%mass_density)

    # compute the volume and mean mass of each cell of the slice
    V_cell = BoxSize_slice**2*depth*1.0/dims**2  #(Mpc/h)^3
    mean_mass = mass_density*V_cell #Msun/h or #

    # compute overdensities
    overdensity /= mean_mass
    print(np.min(overdensity),'< rho/<rho> <',np.max(overdensity))

    # in our convention overdensity(x,y), while for matplotlib is
    # overdensity(y,x), so we need to transpose the field
    overdensity = np.transpose(overdensity)

    # save density field to file
    if save_density_field:  np.save(f_df, overdensity)
    return dx, x, dy, y, overdensity



# code for reading Subfind's subhalo_tab files
# usage e.g.:
#
# import readsubf
# cat = readsubf.subfind_catalog("./m_10002_h_94_501_z3_csf/",63,masstab=True)
# print cat.nsubs
# print "largest halo x position = ",cat.sub_pos[0][0] 

import numpy as np
import os
import sys
 
class subfind_catalog:
  def __init__(self, basedir, snapnum, group_veldisp = False, masstab = False, long_ids = False, swap = False):
    self.filebase = basedir + "/groups_" + str(snapnum).zfill(3) + "/subhalo_tab_" + str(snapnum).zfill(3) + "."
 
    #print
    #print "reading subfind catalog for snapshot",snapnum,"of",basedir
 
    if long_ids: self.id_type = np.uint64
    else: self.id_type = np.uint32
 
    self.group_veldisp = group_veldisp
    self.masstab = masstab
 
    filenum = 0
    doneflag = False
    skip_gr = 0
    skip_sub = 0
    while not doneflag:
      curfile = self.filebase + str(filenum)
      
      if (not os.path.exists(curfile)):
        print("file not found:", curfile)
        sys.exit()
      
      f = open(curfile,'rb')
              
      ngroups = np.fromfile(f, dtype=np.uint32, count=1)[0]
      totngroups = np.fromfile(f, dtype=np.uint32, count=1)[0]
      nids = np.fromfile(f, dtype=np.uint32, count=1)[0]
      totnids = np.fromfile(f, dtype=np.uint64, count=1)[0]
      ntask = np.fromfile(f, dtype=np.uint32, count=1)[0]
      nsubs = np.fromfile(f, dtype=np.uint32, count=1)[0]
      totnsubs = np.fromfile(f, dtype=np.uint32, count=1)[0]
      
      if swap:
        ngroups = ngroups.byteswap()
        totngroups = totngroups.byteswap()
        nids = nids.byteswap()
        totnids = totnids.byteswap()
        ntask = ntask.byteswap()
        nsubs = nsubs.byteswap()
        totnsubs = totnsubs.byteswap()
      
      if filenum == 0:
        self.ngroups = totngroups
        self.nids = totnids
        self.nfiles = ntask
        self.nsubs = totnsubs

        self.group_len = np.empty(totngroups, dtype=np.uint32)
        self.group_offset = np.empty(totngroups, dtype=np.uint32)
        self.group_mass = np.empty(totngroups, dtype=np.float32)
        self.group_pos = np.empty(totngroups, dtype=np.dtype((np.float32,3)))
        self.group_m_mean200 = np.empty(totngroups, dtype=np.float32)
        self.group_r_mean200 = np.empty(totngroups, dtype=np.float32)
        self.group_m_crit200 = np.empty(totngroups, dtype=np.float32)
        self.group_r_crit200 = np.empty(totngroups, dtype=np.float32)
        self.group_m_tophat200 = np.empty(totngroups, dtype=np.float32)
        self.group_r_tophat200 = np.empty(totngroups, dtype=np.float32)
        if group_veldisp:
          self.group_veldisp_mean200 = np.empty(totngroups, dtype=np.float32)
          self.group_veldisp_crit200 = np.empty(totngroups, dtype=np.float32)
          self.group_veldisp_tophat200 = np.empty(totngroups, dtype=np.float32)
        self.group_contamination_count = np.empty(totngroups, dtype=np.uint32)
        self.group_contamination_mass = np.empty(totngroups, dtype=np.float32)
        self.group_nsubs = np.empty(totngroups, dtype=np.uint32)
        self.group_firstsub = np.empty(totngroups, dtype=np.uint32)
        
        self.sub_len = np.empty(totnsubs, dtype=np.uint32)
        self.sub_offset = np.empty(totnsubs, dtype=np.uint32)
        self.sub_parent = np.empty(totnsubs, dtype=np.uint32)
        self.sub_mass = np.empty(totnsubs, dtype=np.float32)
        self.sub_pos = np.empty(totnsubs, dtype=np.dtype((np.float32,3)))
        self.sub_vel = np.empty(totnsubs, dtype=np.dtype((np.float32,3)))
        self.sub_cm = np.empty(totnsubs, dtype=np.dtype((np.float32,3)))
        self.sub_spin = np.empty(totnsubs, dtype=np.dtype((np.float32,3)))
        self.sub_veldisp = np.empty(totnsubs, dtype=np.float32)
        self.sub_vmax = np.empty(totnsubs, dtype=np.float32)
        self.sub_vmaxrad = np.empty(totnsubs, dtype=np.float32)
        self.sub_halfmassrad = np.empty(totnsubs, dtype=np.float32)
        self.sub_id_mostbound = np.empty(totnsubs, dtype=self.id_type)
        self.sub_grnr = np.empty(totnsubs, dtype=np.uint32)
        if masstab:
          self.sub_masstab = np.empty(totnsubs, dtype=np.dtype((np.float32,6)))
     
      if ngroups > 0:
        locs = slice(skip_gr, skip_gr + ngroups)
        self.group_len[locs] = np.fromfile(f, dtype=np.uint32, count=ngroups)
        self.group_offset[locs] = np.fromfile(f, dtype=np.uint32, count=ngroups)
        self.group_mass[locs] = np.fromfile(f, dtype=np.float32, count=ngroups)
        self.group_pos[locs] = np.fromfile(f, dtype=np.dtype((np.float32,3)), count=ngroups)
        self.group_m_mean200[locs] = np.fromfile(f, dtype=np.float32, count=ngroups)
        self.group_r_mean200[locs] = np.fromfile(f, dtype=np.float32, count=ngroups)
        self.group_m_crit200[locs] = np.fromfile(f, dtype=np.float32, count=ngroups)
        self.group_r_crit200[locs] = np.fromfile(f, dtype=np.float32, count=ngroups)
        self.group_m_tophat200[locs] = np.fromfile(f, dtype=np.float32, count=ngroups)
        self.group_r_tophat200[locs] = np.fromfile(f, dtype=np.float32, count=ngroups)
        if group_veldisp:
          self.group_veldisp_mean200[locs] = np.fromfile(f, dtype=np.float32, count=ngroups)
          self.group_veldisp_crit200[locs] = np.fromfile(f, dtype=np.float32, count=ngroups)
          self.group_veldisp_tophat200[locs] = np.fromfile(f, dtype=np.float32, count=ngroups)
        self.group_contamination_count[locs] = np.fromfile(f, dtype=np.uint32, count=ngroups)
        self.group_contamination_mass[locs] = np.fromfile(f, dtype=np.float32, count=ngroups)
        self.group_nsubs[locs] = np.fromfile(f, dtype=np.uint32, count=ngroups)
        self.group_firstsub[locs] = np.fromfile(f, dtype=np.uint32, count=ngroups)        
        skip_gr += ngroups
        
      if nsubs > 0:
        locs = slice(skip_sub, skip_sub + nsubs)
        self.sub_len[locs] = np.fromfile(f, dtype=np.uint32, count=nsubs)
        self.sub_offset[locs] = np.fromfile(f, dtype=np.uint32, count=nsubs)
        self.sub_parent[locs] = np.fromfile(f, dtype=np.uint32, count=nsubs)
        self.sub_mass[locs] = np.fromfile(f, dtype=np.float32, count=nsubs)
        self.sub_pos[locs] = np.fromfile(f, dtype=np.dtype((np.float32,3)), count=nsubs)
        self.sub_vel[locs] = np.fromfile(f, dtype=np.dtype((np.float32,3)), count=nsubs)
        self.sub_cm[locs] = np.fromfile(f, dtype=np.dtype((np.float32,3)), count=nsubs)
        self.sub_spin[locs] = np.fromfile(f, dtype=np.dtype((np.float32,3)), count=nsubs)
        self.sub_veldisp[locs] = np.fromfile(f, dtype=np.float32, count=nsubs)
        self.sub_vmax[locs] = np.fromfile(f, dtype=np.float32, count=nsubs)
        self.sub_vmaxrad[locs] = np.fromfile(f, dtype=np.float32, count=nsubs)
        self.sub_halfmassrad[locs] = np.fromfile(f, dtype=np.float32, count=nsubs)
        self.sub_id_mostbound[locs] = np.fromfile(f, dtype=self.id_type, count=nsubs)
        self.sub_grnr[locs] = np.fromfile(f, dtype=np.uint32, count=nsubs)
        if masstab:
          self.sub_masstab[locs] = np.fromfile(f, dtype=np.dtype((np.float32,6)), count=nsubs)
        skip_sub += nsubs

      curpos = f.tell()
      f.seek(0,os.SEEK_END)
      if curpos != f.tell(): print("Warning: finished reading before EOF for file",filenum)
      f.close()  
      #print 'finished with file number',filenum,"of",ntask
      filenum += 1
      if filenum == self.nfiles: doneflag = True
       
    if swap:
      self.group_len.byteswap(True)
      self.group_offset.byteswap(True)
      self.group_mass.byteswap(True)
      self.group_pos.byteswap(True)
      self.group_m_mean200.byteswap(True)
      self.group_r_mean200.byteswap(True)
      self.group_m_crit200.byteswap(True)
      self.group_r_crit200.byteswap(True)
      self.group_m_tophat200.byteswap(True)
      self.group_r_tophat200.byteswap(True)
      if group_veldisp:
        self.group_veldisp_mean200.byteswap(True)
        self.group_veldisp_crit200.byteswap(True)
        self.group_veldisp_tophat200.byteswap(True)
      self.group_contamination_count.byteswap(True)
      self.group_contamination_mass.byteswap(True)
      self.group_nsubs.byteswap(True)
      self.group_firstsub.byteswap(True)
        
      self.sub_len.byteswap(True)
      self.sub_offset.byteswap(True)
      self.sub_parent.byteswap(True)
      self.sub_mass.byteswap(True)
      self.sub_pos.byteswap(True)
      self.sub_vel.byteswap(True)
      self.sub_cm.byteswap(True)
      self.sub_spin.byteswap(True)
      self.sub_veldisp.byteswap(True)
      self.sub_vmax.byteswap(True)
      self.sub_vmaxrad.byteswap(True)
      self.sub_halfmassrad.byteswap(True)
      self.sub_id_mostbound.byteswap(True)
      self.sub_grnr.byteswap(True)
      if masstab:
        self.sub_masstab.byteswap(True)
       
    #print
    #print "number of groups =", self.ngroups
    #print "number of subgroups =", self.nsubs
    #if self.nsubs > 0:
    #  print "largest group of length",self.group_len[0],"has",self.group_nsubs[0],"subhalos"
    #  print



# code for reading Subfind's ID files
# usage e.g.:
#
# import readsubf
# ids = readsubf.subf_ids("./m_10002_h_94_501_z3_csf/", 0, 100)


class subf_ids:
  def __init__(self, basedir, snapnum, substart, sublen, swap = False, verbose = False, long_ids = False, read_all = False):
    self.filebase = basedir + "/groups_" + str(snapnum).zfill(3) + "/subhalo_ids_" + str(snapnum).zfill(3) + "."

    if (verbose):
      print("reading subhalo IDs for snapshot",snapnum,"of",basedir)
 
    if long_ids: self.id_type = np.uint64
    else: self.id_type = np.uint32

 
    filenum = 0
    doneflag = False
    count=substart
    found=0


    while not doneflag:
      curfile = self.filebase + str(filenum)
      
      if (not os.path.exists(curfile)):
        print("file not found:", curfile)
        sys.exit()
      
      f = open(curfile,'rb')
              
      Ngroups = np.fromfile(f, dtype=np.uint32, count=1)[0]
      TotNgroups = np.fromfile(f, dtype=np.uint32, count=1)[0]
      NIds = np.fromfile(f, dtype=np.uint32, count=1)[0]
      TotNids = np.fromfile(f, dtype=np.uint64, count=1)[0]
      NTask = np.fromfile(f, dtype=np.uint32, count=1)[0]
      Offset = np.fromfile(f, dtype=np.uint32, count=1)[0]


      if read_all:
              substart=0
              sublen=TotNids
      if swap:
          Ngroups = Ngroups.byteswap()
          TotNgroups = TotNgroups.byteswap()
          NIds = NIds.byteswap()
          TotNids = TotNids.byteswap()
          NTask = NTask.byteswap()
          Offset = Offset.byteswap()
      if filenum == 0:
        if (verbose):
          print("Ngroups    = ", Ngroups)
          print("TotNgroups = ", Ngroups)
          print("NIds       = ", NIds)
          print("TotNids    = ", TotNids)
          print("NTask      = ", NTask)
          print("Offset     = ", Offset)
        self.nfiles = NTask
        self.SubLen=sublen
        self.SubIDs = np.empty(sublen, dtype=self.id_type)


      if count <= Offset+NIds:
        nskip = count - Offset
        nrem = Offset + NIds - count
        if sublen > nrem:
          n_to_read = nrem
        else:
          n_to_read = sublen
        if n_to_read > 0:
          if (verbose):
            print(filenum, n_to_read)
          if nskip > 0:
            dummy=np.fromfile(f, dtype=self.id_type, count=nskip)
            if (verbose):
              print(dummy)
          locs = slice(found, found + n_to_read)
          dummy2 = np.fromfile(f, dtype=self.id_type, count=n_to_read)
          if (verbose):
            print(dummy2)
          self.SubIDs[locs]=dummy2
          found += n_to_read
        count += n_to_read
        sublen -= n_to_read

      f.close()
      filenum += 1
      if filenum == self.nfiles: doneflag = True
       
    if swap:
      self.SubIDs.byteswap(True)

 



# This script tests the performance and results of the MASL and CIC libraries
import numpy as np
import MAS_library as MASL
import CIC_library as CIC
import sys,os,time


dims, BoxSize = 256, 1.0
dims2,dims3  = dims**2*1.0, dims**3*1.0

pos_3D = np.random.random((dims**3,3)).astype(np.float32)
pos_2D = np.random.random((dims**2,2)).astype(np.float32)
W      = np.ones(dims**3,dtype=np.float32)
print('positions and weights created')

# do a loop over the different MAS schemes
for MAS in ['NGP','CIC','TSC','PCS']:

    # do a loop over 3D and 2D arrays
    for pos,vals,elements in zip([pos_3D,pos_2D],
                                 [(dims,dims,dims),(dims,dims)],
                                 [dims3,dims2]):

        # do the calculation w/ and w/o weights
        for W in [None,W]:

            delta = np.zeros(vals, dtype=np.float32)
            if W is None:  MASL.MA(pos,delta,BoxSize,MAS)
            else:          MASL.MA(pos,delta,BoxSize,MAS,W)
            if abs(np.sum(delta, dtype=np.float64)/elements-1.0)<1e-5:
                print('Test passed!!!')
            else:
                print('Test not passed!!!');  sys.exit()



#### old CIC library ####
delta = np.zeros(dims**3, dtype=np.float32)
start = time.time()
CIC.NGP_serial(pos_3D,dims,BoxSize,delta)
print('Time taken = %.3f seconds'%(time.time()-start))

delta = np.zeros(dims**3, dtype=np.float32)
start = time.time()
CIC.CIC_serial(pos_3D,dims,BoxSize,delta)
print('Time taken = %.3f seconds'%(time.time()-start))

delta = np.zeros(dims**3, dtype=np.float32)
start = time.time()
CIC.TSC_serial(pos_3D,dims,BoxSize,delta)
print('Time taken = %.3f seconds'%(time.time()-start))
#########################







# import readsnapHDF5 as rs
# header = rs.snapshot_header("snap_063.0") 
# mass = rs.read_block("snap_063","MASS",parttype=5) # reads mass for particles of type 5

import numpy as np
import os
import sys
import math
import tables

############ 
#DATABLOCKS#
############
#descriptions of all datablocks -> add datablocks here!
#TAG:[HDF5_NAME,DIM]
datablocks = {"POS ":["Coordinates",3], 
              "VEL ":["Velocities",3],
              "ID  ":["ParticleIDs",1],
              "MASS":["Masses",1],
              "U   ":["InternalEnergy",1],
              "RHO ":["Density",1],
              "VOL ":["Volume",1],
              "CMCE":["Center-of-Mass",3],
              "AREA":["Surface Area",1],
              "NFAC":["Number of faces of cell",1],
              "NE  ":["ElectronAbundance",1],
              "NH  ":["NeutralHydrogenAbundance",1],
              "HSML":["SmoothingLength",1],
              "SFR ":["StarFormationRate",1],
              "AGE ":["StellarFormationTime",1],
              "Z   ":["Metallicity",1],
              "ACCE":["Acceleration",3],
              "VEVE":["VertexVelocity",3],
              "FACA":["MaxFaceAngle",1],
              "COOR":["CoolingRate",1],
              "POT ":["Potential",1],
              "MACH":["MachNumber",1],
              "GAGE":["GFM StellarFormationTime",1],
              "GIMA":["GFM InitialMass",1],
              "GZ  ":["GFM Metallicity",1],
              "GMET":["GFM Metals",9],
              "GMRE":["GFM MetalsReleased",9],
              "GMAR":["GFM MetalMassReleased", 1]}


#####################################################################################################################
#                                                    READING ROUTINES			                            #
#####################################################################################################################


########################### 
#CLASS FOR SNAPSHOT HEADER#
###########################  
class snapshot_header:
    def __init__(self, *args, **kwargs):
        if (len(args) == 1):
            filename = args[0]
            if os.path.exists(filename):
                curfilename=filename
            elif os.path.exists(filename+".hdf5"):
                curfilename = filename+".hdf5"
            elif os.path.exists(filename+".0.hdf5"): 
                curfilename = filename+".0.hdf5"
            else:
                print("[error] file not found : ", filename)
                sys.exit()

            f=tables.open_file(curfilename)
            self.npart = f.root.Header._v_attrs.NumPart_ThisFile 
            self.nall = f.root.Header._v_attrs.NumPart_Total
            self.nall_highword = f.root.Header._v_attrs.NumPart_Total_HighWord
            self.massarr = f.root.Header._v_attrs.MassTable 
            self.time = f.root.Header._v_attrs.Time 
            self.redshift = f.root.Header._v_attrs.Redshift 
            self.boxsize = f.root.Header._v_attrs.BoxSize
            self.filenum = f.root.Header._v_attrs.NumFilesPerSnapshot
            self.omega0 = f.root.Header._v_attrs.Omega0
            self.omegaL = f.root.Header._v_attrs.OmegaLambda
            self.hubble = f.root.Header._v_attrs.HubbleParam
            self.sfr = f.root.Header._v_attrs.Flag_Sfr 
            self.cooling = f.root.Header._v_attrs.Flag_Cooling
            self.stellar_age = f.root.Header._v_attrs.Flag_StellarAge
            self.metals = f.root.Header._v_attrs.Flag_Metals
            self.feedback = f.root.Header._v_attrs.Flag_Feedback
            self.double = f.root.Header._v_attrs.Flag_DoublePrecision #GADGET-2
            f.close()

        else:
            #read arguments
            self.npart = kwargs.get("npart")
            self.nall = kwargs.get("nall")
            self.nall_highword = kwargs.get("nall_highword")
            self.massarr = kwargs.get("massarr")
            self.time = kwargs.get("time")
            self.redshift = kwargs.get("redshift")
            self.boxsize = kwargs.get("boxsize")
            self.filenum = kwargs.get("filenum")
            self.omega0 = kwargs.get("omega0")
            self.omegaL = kwargs.get("omegaL")
            self.hubble = kwargs.get("hubble")
            self.sfr = kwargs.get("sfr")
            self.cooling = kwargs.get("cooling")
            self.stellar_age = kwargs.get("stellar_age")
            self.metals = kwargs.get("metals")
            self.feedback = kwargs.get("feedback")
            self.double = kwargs.get("double")

            #set default values
            if (self.npart == None):
                self.npart = np.array([0,0,0,0,0,0], dtype="int32")
            if (self.nall == None):
                self.nall  = np.array([0,0,0,0,0,0], dtype="uint32")
            if (self.nall_highword == None):				
                self.nall_highword = np.array([0,0,0,0,0,0], dtype="uint32")
            if (self.massarr == None):
                self.massarr = np.array([0,0,0,0,0,0], dtype="float64")
            if (self.time == None):				
                self.time = np.array([0], dtype="float64")
            if (self.redshift == None):				
                self.redshift = np.array([0], dtype="float64")
            if (self.boxsize == None):				
                self.boxsize = np.array([0], dtype="float64")
            if (self.filenum == None):
                self.filenum = np.array([1], dtype="int32")
            if (self.omega0 == None):
                self.omega0 = np.array([0], dtype="float64")
            if (self.omegaL == None):
                self.omegaL = np.array([0], dtype="float64")
            if (self.hubble == None):
                self.hubble = np.array([0], dtype="float64")
            if (self.sfr == None):	
                self.sfr = np.array([0], dtype="int32")            
            if (self.cooling == None):	
                self.cooling = np.array([0], dtype="int32")
            if (self.stellar_age == None):	
                self.stellar_age = np.array([0], dtype="int32")
            if (self.metals == None):	
                self.metals = np.array([0], dtype="int32")
            if (self.feedback == None):	
                self.feedback = np.array([0], dtype="int32")
            if (self.double == None):
                self.double = np.array([0], dtype="int32")
        
            
        

##############################
#READ ROUTINE FOR SINGLE FILE#
############################## 
def read_block_single_file(filename, block_name, dim2, parttype=-1, no_mass_replicate=False, verbose=False):

  if (verbose):
      print("[single] reading file           : ", filename)   	
      print("[single] reading                : ", block_name)
      
  head = snapshot_header(filename)
  npart = head.npart
  massarr = head.massarr
  nall = head.nall
  filenum = head.filenum
  doubleflag = head.double #GADGET-2
  #doubleflag = 0 #GADGET-2
  del head

  f=tables.open_file(filename)


  #read specific particle type 
  if parttype>=0:
    if (verbose):
        print("[single] parttype               : ", parttype)
    if ((block_name=="Masses") & (npart[parttype]>0) & (massarr[parttype]>0)):
        if (verbose):
            print("[single] replicate mass block")
        ret_val=np.repeat(massarr[parttype], npart[parttype])
    else:
        part_name='PartType'+str(parttype)
        ret_val = f.root._f_get_child(part_name)._f_get_child(block_name)[:]
    if (verbose):
        print("[single] read particles (total) : ", ret_val.shape[0]//dim2)

  #read all particle types
  if parttype==-1:
    first=True
    dim1=0
    for parttype in range(0,5):
        part_name='PartType'+str(parttype)
        if (f.root.__contains__(part_name)):
            if (verbose):
                print("[single] parttype               : ", parttype) 
                print("[single] massarr                : ", massarr)
                print("[single] npart                  : ", npart)

            if ((block_name=="Masses") & (npart[parttype]>0) & (massarr[parttype]>0) & (no_mass_replicate==False)):
                if (verbose):
                    print("[single] replicate mass block")
                if (first):
                    data=np.repeat(massarr[parttype], npart[parttype])
                    dim1+=data.shape[0]
                    ret_val=data
                    first=False
                else:
                    data=np.repeat(massarr[parttype], npart[parttype])
                    dim1+=data.shape[0]
                    ret_val=np.append(ret_val, data)
                if (verbose):
                    print("[single] read particles (total) : ", ret_val.shape[0]//dim2)
                if (doubleflag==0):
                    ret_val=ret_val.astype("float32")
            if (f.root._f_get_child(part_name).__contains__(block_name)):
                if (first):
                    data=f.root._f_get_child(part_name)._f_get_child(block_name)[:]
                    dim1+=data.shape[0]
                    ret_val=data
                    first=False
                else:
                    data=f.root._f_get_child(part_name)._f_get_child(block_name)[:]
                    dim1+=data.shape[0]
                    ret_val=np.append(ret_val, data)
                if (verbose):
                    print("[single] read particles (total) : ", ret_val.shape[0]//dim2)

    if ((dim1>0) & (dim2>1)):
        ret_val=ret_val.reshape(dim1,dim2)

  f.close()

  return ret_val

##############
#READ ROUTINE#
##############
def read_block(filename, block, parttype=-1, no_mass_replicate=False, verbose=False):
  if (verbose):
          print("reading block          : ", block)

  if parttype not in [-1,0,1,2,3,4,5]:
    print("[error] wrong parttype given")
    sys.exit()

  curfilename=filename+".hdf5"

  if os.path.exists(curfilename):
    multiple_files=False
  elif os.path.exists(filename+".0"+".hdf5"):
    curfilename = filename+".0"+".hdf5"
    multiple_files=True
  else:
    print("[error] file not found : ", filename)
    sys.exit()

  head = snapshot_header(curfilename)
  filenum = head.filenum
  del head

  #if (datablocks.has_key(block)):
  if block in datablocks:
        block_name=datablocks[block][0]
        dim2=datablocks[block][1]
        first=True
        if (verbose):
                print("Reading HDF5           : ", block_name)
                print("Data dimension         : ", dim2)
                print("Multiple file          : ", multiple_files)
  else:
        print("[error] Block type ", block, "not known!")
        sys.exit()


  if (multiple_files):
    first=True
    dim1=0
    for num in range(0,filenum):
        curfilename=filename+"."+str(num)+".hdf5"
        if (verbose):
            print("Reading file           : ", num, curfilename)
        if (first):
            data = read_block_single_file(curfilename, block_name, dim2, parttype, verbose)
            dim1+=data.shape[0]
            ret_val = data
            first = False 
        else:
            data = read_block_single_file(curfilename, block_name, dim2, parttype, verbose)
            dim1+=data.shape[0]
            ret_val=np.append(ret_val, data)
        if (verbose):
            print("Read particles (total) : ", ret_val.shape[0]//dim2)

    if ((dim1>0) & (dim2>1)):
        ret_val=ret_val.reshape(dim1,dim2)	
  else:
    ret_val=read_block_single_file(curfilename, block_name, dim2, parttype, no_mass_replicate, verbose)

  return ret_val


#############
#LIST BLOCKS#
#############
def list_blocks(filename, parttype=-1, verbose=False):
  
  f=tables.open_file(filename)
  for parttype in range(0,5):
    part_name='PartType'+str(parttype)
    if (f.root.__contains__(part_name)):
        print("Parttype contains : ", parttype)
        print("-------------------")
        iter = it=datablocks.__iter__()
        next = iter.next()
        while (1):
            if (verbose):
                print("check ", next, datablocks[next][0])
            if (f.root._f_get_child(part_name).__contains__(datablocks[next][0])):
                print(next, datablocks[next][0])
            try:
                next=iter.next()
            except StopIteration:
                break
  f.close()

#################
#CONTAINS BLOCKS#
#################
def contains_block(filename, tag, parttype=-1, verbose=False):
  
  contains_flag=False
  f=tables.open_file(filename)
  for parttype in range(0,5):
        part_name='PartType'+str(parttype)
        if (f.root.__contains__(part_name)):
                iter = it=datablocks.__iter__()
                next = iter.next()
                while (1):
                        if (verbose):
                                print("check ", next, datablocks[next][0])
                        if (f.root._f_get_child(part_name).__contains__(datablocks[next][0])):
                            if (next.find(tag)>-1):
                                contains_flag=True
                        try:
                            next=iter.next()
                        except StopIteration:
                            break
  f.close() 
  return contains_flag

############
#CHECK FILE#
############
def check_file(filename):
  f=tables.open_file(filename)
  f.close()
                                                                                                                                                  






#####################################################################################################################
#                                                    WRITING ROUTINES    		                            #
#####################################################################################################################



#######################
#OPEN FILE FOR WRITING#
#######################
def openfile(filename):
    f=tables.open_file(filename, mode = "w")
    return f

############
#CLOSE FILE#
############
def closefile(f):
    f.close()

##############################
#WRITE SNAPSHOT HEADER OBJECT#
##############################
def writeheader(f, header):	
    group_header=f.createGroup(f.root, "Header")
    group_header._v_attrs.NumPart_ThisFile=header.npart
    group_header._v_attrs.NumPart_Total=header.nall
    group_header._v_attrs.NumPart_Total_HighWord=header.nall_highword
    group_header._v_attrs.MassTable=header.massarr
    group_header._v_attrs.Time=header.time
    group_header._v_attrs.Redshift=header.redshift
    group_header._v_attrs.BoxSize=header.boxsize
    group_header._v_attrs.NumFilesPerSnapshot=header.filenum
    group_header._v_attrs.Omega0=header.omega0
    group_header._v_attrs.OmegaLambda=header.omegaL
    group_header._v_attrs.HubbleParam=header.hubble
    group_header._v_attrs.Flag_Sfr=header.sfr
    group_header._v_attrs.Flag_Cooling=header.cooling
    group_header._v_attrs.Flag_StellarAge=header.stellar_age
    group_header._v_attrs.Flag_Metals=header.metals
    group_header._v_attrs.Flag_Feedback=header.feedback
    group_header._v_attrs.Flag_DoublePrecision=header.double

###############
#WRITE ROUTINE#
###############
def write_block(f, block, parttype, data):
    part_name="PartType"+str(parttype)
    if (f.root.__contains__(part_name)==False):
            group=f.createGroup(f.root, part_name)
    else:
        group=f.root._f_get_child(part_name)
    
    #if (datablocks.has_key(block)):
    if block in datablocks:
        block_name=datablocks[block][0]
        dim2=datablocks[block][1]
        if (group.__contains__(block_name)==False):
            table=f.createArray(group, block_name, data)
        else:
            print("I/O block already written")
    else:
        print("Unknown I/O block")



from setuptools import setup
from setuptools.extension import Extension
from Cython.Distutils import build_ext
from Cython.Build import cythonize
import numpy


setup(name = 'c_wrapper',
      cmdclass = {'build_ext': build_ext},
      ext_modules = cythonize([Extension("cython_code", 
        ["cython_code.pyx",'c_code.c'],
        compiler_directives={'language_level' : "3"},
        include_dirs=[numpy.get_include()],
        extra_compile_args=["-O3","-ffast-math","-march=native"])]))



#This library is used for the HI & clusters project

import numpy as np
import readsnap,readsnap2
import sys,os,time 
cimport numpy as np

################################# UNITS #####################################
rho_crit = 2.77536627e11 #h^2 Msun/Mpc^3

yr    = 3.15576e7  #seconds
km    = 1e5        #cm
Mpc   = 3.0856e24  #cm
kpc   = 3.0856e21  #cm
Msun  = 1.989e33   #g
Ymass = 0.24       #helium mass fraction
mH    = 1.6726e-24 #proton mass in grams
gamma = 5.0/3.0    #ideal gas
kB    = 1.3806e-26 #gr (km/s)^2 K^{-1}
nu0   = 1420.0     #21-cm frequency in MHz

pi = np.pi
#############################################################################

################################ Rahmati ######################################
#Notice that in this implementation we set the temperature of the star-forming
#particles to 1e4 K, and use the Rahmati formula to compute the HI/H fraction.
#The routine computes and returns the HI/H fraction taking into account the
#HI self-shielding and the presence of molecular hydrogen.
#snapshot_fname ----------> file containing the snapshot
#TREECOOL_file -----------> file containing the TREECOOL file
#T_block -----------------> whether the snapshot contains the TEMP block or not
#Gamma_UVB ---------------> value of Gamma_UVB. None to read it from TREECOOL
#SF_temperature ----------> The value to set the temperature of the SF particles
#self_shielding_correction --> whether to correct HI fraction for self-shielding
#correct_H2 --------------> correct HI fraction for presence of H2:
#                           'BR': Blitz-Rosolowsky, 'THINGS', 'KMT' or 'None'
def Rahmati(snapshot_fname, TREECOOL_file, T_block=True, Gamma_UVB=None,
            SF_temperature=1e4, self_shielding_correction=True,
            correct_H2='BR'):

    # read snapshot head and obtain BoxSize, Omega_m and Omega_L 
    head     = readsnap.snapshot_header(snapshot_fname)
    BoxSize  = head.boxsize/1e3 #Mpc/h   
    Nall     = head.nall
    Masses   = head.massarr*1e10 #Msun/h    
    Omega_m  = head.omega_m
    Omega_l  = head.omega_l
    redshift = head.redshift
    Hubble   = 100.0*np.sqrt(Omega_m*(1.0+redshift)**3+Omega_l)  #h*km/s/Mpc 
    h        = head.hubble

    # read/compute the temperature of the gas particles
    if T_block:
        T = readsnap.read_block(snapshot_fname,"TEMP",parttype=0) #K
    else:
        U = readsnap.read_block(snapshot_fname,"U   ",parttype=0) #km/s
        T = U/1.5/(1.380658e-16/(0.6*1.672631e-24)*1e-10)         #K
    T = T.astype(np.float64) #to increase precision use float64 variables
    print('%.3e < T[K] < %.3e'%(np.min(T),np.max(T)))

    # read the density block: units of h^2 Msun/Mpc^3
    rho = readsnap.read_block(snapshot_fname,"RHO ",parttype=0)*1e10/1e-9 
    print('%.3e < rho < %.3e'%(np.min(rho),np.max(rho)))

    # Rahmati et. al. 2013 self-shielding parameters (table A1)
    z_t       = np.array([0.00, 1.00, 2.00, 3.00, 4.00, 5.00])
    n0_t      = np.array([-2.94,-2.29,-2.06,-2.13,-2.23,-2.35]); n0_t=10**n0_t
    alpha_1_t = np.array([-3.98,-2.94,-2.22,-1.99,-2.05,-2.63])
    alpha_2_t = np.array([-1.09,-0.90,-1.09,-0.88,-0.75,-0.57])
    beta_t    = np.array([1.29, 1.21, 1.75, 1.72, 1.93, 1.77])
    f_t       = np.array([0.01, 0.03, 0.03, 0.04, 0.02, 0.01])

    # compute the self-shielding parameters at the redshift of the N-body
    n0      = np.interp(redshift,z_t,n0_t)
    alpha_1 = np.interp(redshift,z_t,alpha_1_t)
    alpha_2 = np.interp(redshift,z_t,alpha_2_t)
    beta    = np.interp(redshift,z_t,beta_t)
    f       = np.interp(redshift,z_t,f_t)
    print('n0 = %e\nalpha_1 = %2.3f\nalpha_2 = %2.3f\nbeta = %2.3f\nf = %2.3f'\
        %(n0,alpha_1,alpha_2,beta,f))

    # find the value of the photoionization rate
    if Gamma_UVB==None:
        data = np.loadtxt(TREECOOL_file);  logz = data[:,0]; 
        Gamma_UVB = data[:,1]
        Gamma_UVB = np.interp(np.log10(1.0+redshift),logz,Gamma_UVB);  del data
        print('Gamma_UVB(z=%2.2f) = %e s^{-1}'%(redshift,Gamma_UVB))


    # for star forming particle assign T=10^4 K
    if SF_temperature!=None and T_block==True:
        SFR = readsnap.read_block(snapshot_fname,"SFR ",parttype=0) #SFR
        indexes = np.where(SFR>0.0)[0];  T[indexes] = SF_temperature
        del indexes,SFR; print('%.3e < T[K] < %.3e'%(np.min(T),np.max(T)))
    

    #### Now compute the HI/H fraction following Rahmati et. al. 2013 ####
    # compute densities in cm^{-3}. rho is in h^2 Msun/Mpc^3 units
    nH = 0.76*h**2*Msun/Mpc**3/mH*rho*(1.0+redshift)**3; #del rho
    nH = nH.astype(np.float64)

    #compute the case A recombination rate
    Lambda  = 315614.0/T
    alpha_A = 1.269e-13*Lambda**(1.503)\
        /(1.0+(Lambda/0.522)**(0.47))**(1.923) #cm^3/s
    alpha_A = alpha_A.astype(np.float64)

    #Compute Lambda_T (eq. A6 of Rahmati et. al. 2013)
    Lambda_T=1.17e-10*np.sqrt(T)*np.exp(-157809.0/T)/(1.0+np.sqrt(T/1e5))#cm^3/s
    Lambda_T=Lambda_T.astype(np.float64)

    #compute the photoionization rate
    Gamma_phot = Gamma_UVB
    if self_shielding_correction:
        Gamma_phot *=\
            ((1.0-f)*(1.0+(nH/n0)**beta)**alpha_1 + f*(1.0+nH/n0)**alpha_2)
    print('Gamma_phot = ',Gamma_phot)

    #compute the coeficients A,B and C to calculate the HI/H fraction
    A = alpha_A + Lambda_T
    B = 2.0*alpha_A + Gamma_phot/nH + Lambda_T
    C = alpha_A

    #compute the HI/H fraction (eq. A8 of Rahmati et. al. 2013)
    nH0 = (B-np.sqrt(B**2-4.0*A*C))/(2.0*A); del nH
    nH0 = nH0.astype(np.float32)

    #correct for the presence of H2
    if correct_H2 in ['BR','THINGS']:
        print('correcting HI/H to account for the presence of H2...')
        #compute the pression of the gas particles
        #h^2Msun/kpc^3
        rho = readsnap.read_block(snapshot_fname,"RHO ",parttype=0)*1e10
        U   = readsnap.read_block(snapshot_fname,"U   ",parttype=0) #(km/s)^2
        P   = (gamma-1.0)*U*rho*(1.0+redshift)**3 #h^2 Msun/kpc^3*(km/s)^2
        P   = h**2*Msun/kpc**3*P                  #gr/cm^3*(km/s)^2
        P  /= kB                                 #K/cm^3
        del rho,U
 
        #assign H2 only to star forming particles
        SFR = readsnap.read_block(snapshot_fname,"SFR ",parttype=0)
        indexes = np.where(SFR>0.0)[0]; del SFR

        #compute the H2/HI fraction
        R_surf = np.zeros(Nall[0],dtype=np.float32)
        if correct_H2   == 'BR':       R_surf = (P/3.5e4)**0.92
        elif correct_H2 == 'THINGS':   R_surf = (P/1.7e4)**0.8
        #R_surf[IDs]=1.0/(1.0+(35.0*(0.1/nH/0.76)**(gamma))**0.92)
        else:
            print('bad choice of correct_H2!!!'); sys.exit()

        #compute the corrected HI mass taking into account the H2
        nH0[indexes]=nH0[indexes]/(1.0+R_surf[indexes]); del indexes,R_surf
        #M_HI[indexes]=M_HI[indexes]*(1.0-R_surf[indexes]); del indexes,R_surf

    if correct_H2=='KMT':
        print('correcting HI/H to account for the presence of H2 using KMT...')
        H2_frac = H2_fraction(snapshot_fname)  #H2/NH = H2/(H2+HI)
        nH0 = nH0*(1.0-H2_frac)                #(NH/H)*(HI/NH)
        if np.any(nH0<0.0):
            print('HI/H cant be negative!!!!'); sys.exit()

    return nH0
##############################################################################

##############################################################################
#This routine computes and returns the metallicity of each gas particle,
#in units of the solar metallicity
def gas_metallicity(snapshot_fname):

    # read snapshot head and obtain BoxSize, Omega_m and Omega_L
    print('finding the metallicity of the gas particles')
    head  = readsnap.snapshot_header(snapshot_fname)
    Nall  = head.nall
    files = head.filenum

    # read the masses of the gas particles in 1e10 Msun/h
    mass = readsnap.read_block(snapshot_fname,"MASS",parttype=0)

    # define the metallicity array
    metallicity = np.zeros(Nall[0],dtype=np.float64)

    offset = 0
    for filenum in range(files):
        if files==1:  fname = snapshot_fname
        else:         fname = snapshot_fname+'.'+str(filenum)
        
        npart = readsnap.snapshot_header(fname).npart  #particles in that file

        block_found = False;  EoF = False;  f = open(fname,'rb')
        f.seek(0,2); last_position = f.tell(); f.seek(0,0)
        while (not(block_found) or not(EoF)):
        
            # read the three first elements and the blocksize
            format_type = np.fromfile(f,dtype=np.int32,count=1)[0]
            block_name  = f.read(4)
            dummy       = np.fromfile(f,dtype=np.float64,count=1)[0] 
            blocksize1  = np.fromfile(f,dtype=np.int32,count=1)[0]
            
            if block_name=='Zs  ':
                Z = np.fromfile(f,dtype=np.float32,count=npart[0]*15)
                Z = np.reshape(Z,(npart[0],15))
            
                metal = np.sum(Z[:,1:],axis=1,dtype=np.float64)\
                    /(mass[offset:offset+npart[0]])/0.0127
                metallicity[offset:offset+npart[0]] = metal;  offset += npart[0]

                f.seek(npart[4]*15*4,1);  block_found = True

            else:   f.seek(blocksize1,1)

            # read the size of the block and check that is the same number
            blocksize2 = np.fromfile(f,dtype=np.int32,count=1)[0]

            if blocksize2!=blocksize1:
                print('error!!!'); sys.exit()
        
            current_position = f.tell()
            if current_position == last_position:  EoF = True
        f.close()

    if offset!=Nall[0]:
        print('Not all files read!!!'); sys.exit()
        
    return metallicity
##############################################################################

##############################################################################
#This routine computes the H2 fraction in the gas particles. Only H2 is assigned
#to star-forming particles. We use the KMT formalism, see also Dave et al. 2013
def H2_fraction(snapshot_fname):

    # read snapshot header
    print('\nREADING SNAPSHOTS PROPERTIES')
    head     = readsnap.snapshot_header(snapshot_fname)
    Nall     = head.nall
    redshift = head.redshift
    h        = head.hubble

    # find the metallicity of the gas particles
    Z = gas_metallicity(snapshot_fname)

    Msun    = 1.99e33   #g
    kpc     = 3.0857e21 #cm
    sigma_d = Z*1e-21   #cm^2
    mu_H    = 2.3e-24   #g

    # read the density of the gas particles in h^2 Msun/kpc^3 in proper units
    rho = readsnap.read_block(snapshot_fname,"RHO ",parttype=0)*1e10 
    rho = rho*h**2*(1.0+redshift)**3  #Msun/kpc^3
 
    # read the SPH radii of the gas particles in kpc/h in proper units
    R = readsnap.read_block(snapshot_fname,"HSML",parttype=0) #ckpc/h
    R = (R/h)/(1.0+redshift)                                  #kpc

    # compute the gas surface density in g/cm^2
    sigma = (rho*R)*(Msun/kpc**2)  #g/cm^2
    del rho, R

    # read the star formation rate. Only star-forming particles will host H2
    SFR = readsnap.read_block(snapshot_fname,"SFR ",parttype=0)
    indexes = np.where((SFR>0.0) & (Z>0))[0]
    
    #define the molecular hydrogen fraction array
    f_H2 = np.zeros(Nall[0],dtype=np.float64)

    # compute the value of tau_c, chi and s of the star-forming particles
    tau_c = sigma[indexes]*(sigma_d[indexes]/mu_H)
    chi   = 0.756*(1.0 + 3.1*Z[indexes]**0.365)
    s     = np.log(1.0 + 0.6*chi + 0.01*chi**2)/(0.6*tau_c)

    f_H2[indexes] = 1.0 - 0.75*s/(1.0 + 0.25*s);  del indexes, tau_c, chi, s
    f_H2[np.where(f_H2<0.0)[0]] = 0.0 #avoid negative values in the H2_fraction

    return f_H2
##############################################################################


# Returns the snapshot_fname and halo_file for a given region and physics
class Region:
    def __init__(self,region,physics,snap_number):

        phys_dict   = {'CSF':'/CSF2015/', 'AGN':'/BH2015/'}
        root_folder = '/pico/home/userexternal/fvillaes/Dianoga/D'+\
            str(region)+phys_dict[physics]

        snapshot_fname = root_folder+'snap_'+snap_number
        halo_file      = root_folder+\
            'PostProcessing/PostProc_OldSetUp/snap_'+snap_number+'.glob'
        subfind_file   = root_folder+'Subfind/groups_'+snap_number+\
            '/sub_'+snap_number

        self.snapshot_fname = snapshot_fname
        self.halo_file      = halo_file
        self.subfind_file   = subfind_file



# this routine renumber the IDs
# IDs ------------> array with the "wrong" IDs to renumber
# bad_IDs --------> array with the sorted bad IDs of the simulation
# missing_IDs ----> array with the missed good IDs of the entire simulation
# there is a one-to-one correspondence between bad_IDs and missing_IDs,
# such as the ID to assign to a particle with a "wrong" ID bad_IDs[j] would be
# missing_IDs[j]. The routine, given a wrong ID, will look for the index j such
# as bad_IDs[j]=ID, and then assign a new ID which would be missing_IDs[j]
def rename_IDs(np.ndarray[np.uint32_t, ndim=1] IDs,
               np.ndarray[np.uint32_t, ndim=1] bad_IDs,
               np.ndarray[np.uint32_t, ndim=1] missing_IDs):

    cdef long i, j_min, j_max, j, value
    cdef np.ndarray[np.uint32_t, ndim=1] new_IDs

    length  = len(IDs);  length_bad = len(bad_IDs)
    new_IDs = np.zeros(length,dtype=IDs.dtype)  #define the new_IDs array

    for i in range(length):
        value = IDs[i]
        j_min = 0;  j_max = length_bad;  j = (j_min+j_max)//2
        while(value!=bad_IDs[j]):
            if value>bad_IDs[j]:   j_min = j
            else:                  j_max = j
            j = (j_min+j_max)//2
        new_IDs[i] = missing_IDs[j]

    return new_IDs


# this routine checks whether there is any repeated ID in the subfind file
# if there are repeated IDs it means that some particles are counted twice
# IDs ------------> array with the IDs to check
def repeated_IDs(np.ndarray[np.uint32_t, ndim=1] IDs):
    
    cdef np.ndarray[np.uint64_t, ndim=1] array
    cdef int i

    array = np.zeros(np.max(IDs)+1,dtype=np.uint64)

    for i in range(len(IDs)):
        array[IDs[i]] += 1

    if np.any(array>1):
        print('some particle is counted twice!!!\n');  sys.exit()
    else:
        print('no repeated IDs found\n')



# This routine takes all halos of a given region and find the galaxies within
# it. It also computes the HI mass of the particles belonging to those galaxies
# within R200 of the halo
# physics -----------------> 'CSF' or 'AGN'
# reg ---------------------> region over which perform the analysis
# TREECOOL_file -----------> TREECOOL file
def HI_in_galaxies(physics,reg,TREECOOL_file):

    #################################################
    # from the region obtain the snapshot fname and the halo file
    region         = Region(reg,physics)
    snapshot_fname = region.snapshot_fname
    halo_file      = region.halo_file
    subfind_file   = region.subfind_file

    # read halo file 
    data  = np.loadtxt(halo_file)
    pos_h = data[:,1:4]/1e3;  R_h  = data[:,17]/1e3;  R200    = data[:,24]/1e3
    M200  = data[:,18];       M500 = data[:,19];      M2500   = data[:,20]
    M_g   = data[:,13];       M_hg = data[:,14];      M_stars = data[:,15] 
    M_cg  = data[:,16];       contaminants = data[:,37]
    if np.any(contaminants!=0):
        print('contaminants founds!!!',contaminants);  sys.exit()
    #################################################

    #################################################
    # find the total number of particles in the simulation
    Nall   = readsnap.snapshot_header(snapshot_fname).nall
    Ntotal = np.sum(Nall,dtype=np.uint64)
    print('Total number of particles in the simulation =',Ntotal)

    # create a zero array and fill it with 1 using the normal IDs
    # We identify the missing IDs as the positions where the array is 0
    IDs_sorted = np.zeros(Ntotal,dtype=np.uint32)

    # find the IDs missing
    IDs = readsnap.read_block(snapshot_fname,"ID  ",parttype=-1)-1 #normalized
    #indexes = np.where((IDs>2150078190) & (IDs<2150078220))[0]
    #print IDs[indexes]
    print('Number of IDs =',len(IDs));  print(np.min(IDs),'< ID <',np.max(IDs))
    IDs_sorted[ IDs[np.where(IDs<Ntotal)[0]] ] = 1 #put 1 in normal ID positions
    missing_IDs = (np.where(IDs_sorted==0)[0]).astype(np.uint32)
    del IDs_sorted;  bad_IDs = IDs[np.where(IDs>=Ntotal)[0]];  del IDs
    #print 'bad IDs =',bad_IDs;  print 'missing IDs =',missing_IDs
    print('Number of particles with IDs longer than Ntotal =',len(missing_IDs))

    # sort the bad IDs
    bad_IDs = np.sort(bad_IDs);  print('bad IDs sorted =',bad_IDs,'\n')
    # we renumber the IDs, such as:
    # bad_IDs[0] --> missing_IDs[0]
    # bad_IDs[1] --> missing_IDs[1] ...etc
    #################################################

    #################################################
    # read subfind file
    groups_pos        = readsnap2.read_block(subfind_file,'GPOS')/1e3  #Mpc/h
    groups_M200       = readsnap2.read_block(subfind_file,'MCRI')*1e10 #Msun/h
    groups_R200       = readsnap2.read_block(subfind_file,'RCRI')/1e3  #Mpc/h
    #groups_M200       = readsnap2.read_block(subfind_file,'M200')*1e10 #Msun/h
    #groups_Mtot       = readsnap2.read_block(subfind_file,'MTOT')*1e10 #Msun/h
    galaxies_per_halo = readsnap2.read_block(subfind_file,'NSUB')
    galaxy_pos        = readsnap2.read_block(subfind_file,'SPOS')/1e3  #Mpc/h
    galaxy_number     = readsnap2.read_block(subfind_file,'SLEN')
    galaxy_offset     = readsnap2.read_block(subfind_file,'SOFF')
    galaxy_IDs        = readsnap2.read_block(subfind_file,'PID ')-1 #Normalized
    galaxy_mass       = readsnap2.read_block(subfind_file,'MSUB')*1e10  #Msun/h

    # renumber the IDs of the particles belonging to subhalos
    bad_IDs_gal_indexes = np.where(galaxy_IDs>=Ntotal)[0]
    bad_IDs_gal         = galaxy_IDs[bad_IDs_gal_indexes]
    if len(bad_IDs_gal)>0:
        galaxy_IDs[bad_IDs_gal_indexes] = \
            rename_IDs(bad_IDs_gal,bad_IDs,missing_IDs)
    del bad_IDs_gal_indexes, bad_IDs_gal
    print(np.min(galaxy_IDs),'< ID galaxy particles <',np.max(galaxy_IDs))

    # check that no IDs are repeated
    repeated_IDs(galaxy_IDs)
    #################################################

    #################################################
    # read the positions, masses and IDs of all the particles
    pos_part  = readsnap.read_block(snapshot_fname,"POS ",parttype=-1)/1e3  
    IDs_part  = readsnap.read_block(snapshot_fname,"ID  ",parttype=-1)-1 
    mass_part = readsnap.read_block(snapshot_fname,"MASS",parttype=-1)*1e10 

    # renumber the wrong IDs
    bad_IDs_part_indexes = np.where(IDs_part>=Ntotal)[0]
    bad_IDs_part         = IDs_part[bad_IDs_part_indexes]
    if len(bad_IDs_part)>0:
        IDs_part[bad_IDs_part_indexes] = \
            rename_IDs(bad_IDs_part, bad_IDs, missing_IDs)
    del bad_IDs_part_indexes, bad_IDs_part

    # sort the positions of the particles
    pos_part_sorted = np.empty((Ntotal,3),dtype=np.float32)
    pos_part_sorted[IDs_part] = pos_part;  del pos_part

    # sort the masses of the particles
    mass_part_sorted = np.empty(Ntotal,dtype=np.float32)
    mass_part_sorted[IDs_part] = mass_part;  del mass_part, IDs_part
    #################################################

    #################################################
    # compute the HI masses of the gas particles
    HI_frac = Rahmati(snapshot_fname, TREECOOL_file, T_block=True,  #HI/H
                           Gamma_UVB=None, SF_temperature=1e4,
                           self_shielding_correction=True, correct_H2='KMT')
    mass_g = readsnap.read_block(snapshot_fname,"MASS",parttype=0)*1e10  #Msun/h
    pos_g  = readsnap.read_block(snapshot_fname,"POS ",parttype=0)/1e3  #Mpc/h
    M_HI   = 0.76*mass_g*HI_frac;  del HI_frac

    # read and renumber the IDs of the gas particles
    IDs_g = readsnap.read_block(snapshot_fname,"ID  ",parttype=0)-1  #normalized
    bad_IDs_g_indexes = np.where(IDs_g>=Ntotal)[0]
    bad_IDs_g         = IDs_g[bad_IDs_g_indexes]
    if len(bad_IDs_g)>0:
        IDs_g[bad_IDs_g_indexes] = rename_IDs(bad_IDs_g,bad_IDs,missing_IDs)
    del bad_IDs_g_indexes, bad_IDs_g

    # create a sorted array with the HI masses
    M_HI_sorted = np.zeros(Ntotal,dtype=np.float64)
    M_HI_sorted[IDs_g] = M_HI

    gas_mass = np.zeros(Ntotal,dtype=np.float64)
    gas_mass[IDs_g] = mass_g;  del IDs_g, mass_g
    #################################################

    #################################################
    # read the stellar masses of the particles
    mass_s = readsnap.read_block(snapshot_fname,"MASS",parttype=4)*1e10  #Msun/h

    # read and renumber the IDs of the star particles
    IDs_s  = readsnap.read_block(snapshot_fname,"ID  ",parttype=4)-1 #normalized
    bad_IDs_s_indexes = np.where(IDs_s>=Ntotal)[0]
    bad_IDs_s         = IDs_s[bad_IDs_s_indexes]
    if len(bad_IDs_s)>0:
        IDs_s[bad_IDs_s_indexes] = rename_IDs(bad_IDs_s,bad_IDs,missing_IDs)
    del bad_IDs_s_indexes, bad_IDs_s

    # create a sorted array with the stellar masses
    M_stars_sorted = np.zeros(Ntotal,dtype=np.float64)
    M_stars_sorted[IDs_s] = mass_s;  del IDs_s, mass_s
    #################################################
    

    #################################################
    # do a loop over all halos
    M_HI_halos          = np.zeros(len(R200),dtype=np.float64)
    M_HI_galaxies_halos = np.zeros(len(R200),dtype=np.float64)
    galaxies            = np.zeros(len(R200),dtype=np.int32) 
    galaxies_with_HI    = np.zeros(len(R200),dtype=np.int32) 
    galaxies_with_stars = np.zeros(len(R200),dtype=np.int32) 
    
    for i in range(len(R200)):

        ######## compute the HI mass of the halo ########
        distance = np.sqrt((pos_g[:,0]-pos_h[i,0])**2 + \
                           (pos_g[:,1]-pos_h[i,1])**2 + \
                           (pos_g[:,2]-pos_h[i,2])**2)

        indexes  = np.where(distance<R200[i])[0]
        HI_mass  = np.sum(M_HI[indexes],dtype=np.float64)
        del indexes, distance;  print('HI mass = %.4e'%HI_mass)
        #################################################

        ######## find the galaxies within R200 ########
        distance = galaxy_pos - pos_h[i]
        distance = np.sqrt(distance[:,0]**2+distance[:,1]**2+distance[:,2]**2)
        indexes  = np.where(distance<R200[i])[0]        
        number_galaxies = len(indexes);  del distance
        print('galaxies found within the halo:',number_galaxies)
        ###############################################

        ###### find the IDs particles in the galaxies within R200 ######
        length = 0   #find the total number of particles
        for j in indexes:  length += galaxy_number[j]
        print('particles belonging to galaxies within R200:',length)
        
        IDs_part_gal = np.empty(length,dtype=np.uint32);  offset = 0
        for j in indexes:
            size_gal     = galaxy_number[j]
            offset_gal   = galaxy_offset[j]
            gal_IDs      = galaxy_IDs[offset_gal:offset_gal+size_gal] 
            HI_galaxy    = np.sum(M_HI_sorted[gal_IDs],dtype=np.float64)
            mass_galaxy  = np.sum(mass_part_sorted[gal_IDs],dtype=np.float64)
            stellar_mass = np.sum(M_stars_sorted[gal_IDs],dtype=np.float64)
            gas_galaxy   = np.sum(gas_mass[gal_IDs],dtype=np.float64)
            if HI_galaxy>0.0:   galaxies_with_HI[i] += 1
            if stellar_mass>0:  galaxies_with_stars[i] += 1
            #if galaxy_pos[j,0]>502.7 and galaxy_pos[j,0]<503.2 and \
            #        galaxy_pos[j,1]>497.1 and galaxy_pos[j,1]<497.3:
            print('%3d ---> %.4e %.4e : %.4e %.4e %.4e'\
                %(j,mass_galaxy,galaxy_mass[j],gas_galaxy,
                  HI_galaxy,stellar_mass))
                #print galaxy_pos[j]
            #if np.absolute(mass_galaxy-galaxy_mass[j])/mass_galaxy>1e-5:
            #    sys.exit()
            IDs_part_gal[offset:offset+size_gal] = gal_IDs
            offset += size_gal
        del length, indexes
        ################################################################

        # find the positions of the particles within galaxies in R200
        gal_part_pos = pos_part_sorted[IDs_part_gal]

        # select the particles belonging to galaxies within R200 of the halo
        distance = gal_part_pos - pos_h[i]
        distance = np.sqrt(distance[:,0]**2+distance[:,1]**2+distance[:,2]**2)
        indexes = np.where(distance<R200[i])[0]  
        gal_part_pos = gal_part_pos[indexes]
        IDs_part_gal = IDs_part_gal[indexes]
        HI_mass_galaxies = np.sum(M_HI_sorted[IDs_part_gal],dtype=np.float64)
        print('HI mass in galaxies within R200 = %.4e'\
            %HI_mass_galaxies)


        M_HI_halos[i]          = HI_mass
        M_HI_galaxies_halos[i] = HI_mass_galaxies
        galaxies[i]            = number_galaxies

    return M200, M_HI_halos, M_HI_galaxies_halos, galaxies, galaxies_with_HI,\
        galaxies_with_stars, pos_h
